/*
 * linux/arch/arm/mach-at91rm9200/board-matrix500.c
 *
 *  Copyright (C) 2005 SAN People
 *
 *  Epson S1D framebuffer glue code is:
 *     Copyright (C) 2005 Thibaut VARENE <varenet@parisc-linux.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/device.h>

#include <asm/hardware.h>
#include <asm/setup.h>
#include <asm/mach-types.h>
#include <asm/irq.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>

#include <asm/arch/hardware.h>
#include <asm/mach/serial_at91rm9200.h>
#include <asm/arch/board.h>

#include "generic.h"

static void __init matrix500_init_irq(void)
{
	/* Initialize AIC controller */
	at91rm9200_init_irq(NULL);

	/* Set up the GPIO interrupts */
	at91_gpio_irq_setup(BGA_GPIO_BANKS);
}

/*
 * Serial port configuration.
 *    0 .. 3 = USART0 .. USART3
 *    4      = DBGU
 */
#define MATRIX500_UART_MAP		{ 4, 0, 1, 2, 3 }	/* ttyS0(debug),ttyS1(USART0) ..., ttyS4(USART3) */
#define MATRIX500_SERIAL_CONSOLE	0			/* ttyS0 */

static void __init matrix500_map_io(void)
{
	int serial[AT91C_NR_UART] = MATRIX500_UART_MAP;
	int i;

	at91rm9200_map_io();

	/* Initialize clocks: 18.432 MHz crystal */
	at91_clock_init(18432000);

#ifdef CONFIG_SERIAL_AT91
	at91_console_port = MATRIX500_SERIAL_CONSOLE;
	memcpy(at91_serial_map, serial, sizeof(serial));
	
	/* Register UARTs */
	for (i = 0; i < AT91C_NR_UART; i++) {
		if (at91_serial_map[i] >= 0)
			at91_register_uart(i, at91_serial_map[i]);
	}
#endif
}


static struct at91_eth_data __initdata matrix500_eth_data = {
	.phy_irq_pin	= AT91_PIN_PC4,
	.is_rmii	= 1,
};

static struct at91_usbh_data __initdata matrix500_usbh_data = {
	.ports		= 2,
};

static struct at91_udc_data __initdata matrix500_udc_data = {
	.vbus_pin	= AT91_PIN_PD4,
	.pullup_pin	= AT91_PIN_PD5,
};

static struct at91_mmc_data __initdata matrix500_mmc_data = {
	.is_b		= 0,
	.wire4		= 1,
	.det_pin	= AT91_PIN_PD19,//alan
};

static void __init matrix500_board_init(void)
{
	/* Ethernet */
	at91_add_device_eth(&matrix500_eth_data);
	//at91_add_device_dm9000();
	/* USB Host */
	at91_add_device_usbh(&matrix500_usbh_data);
	/* USB Device */
	at91_add_device_udc(&matrix500_udc_data);
	/* MMC Device */
	at91_add_device_mmc(&matrix500_mmc_data);

	/* alan: Power On debug LED 4 */
	at91_set_gpio_output(AT91_PIN_PD0, 1);
	at91_set_gpio_output(AT91_PIN_PD1, 1);
	at91_set_gpio_output(AT91_PIN_PD2, 1);
	at91_set_gpio_output(AT91_PIN_PD3, 1);
	/* alan: Power On Ready LED */
	at91_set_gpio_output(AT91_PIN_PD6, 0);
}

MACHINE_START(M501, "Artila M501 SOM")
	/* Maintainer: SAN People/Atmel */
	.phys_ram	= AT91_SDRAM_BASE,
	.phys_io	= AT91C_BASE_SYS,
	.io_pg_offst	= (AT91C_VA_BASE_SYS >> 18) & 0xfffc,
	.boot_params	= AT91_SDRAM_BASE + 0x100,
	.timer		= &at91rm9200_timer,
	.map_io		= matrix500_map_io,
	.init_irq	= matrix500_init_irq,
	.init_machine	= matrix500_board_init,
MACHINE_END
