/*
 * Artila GPIO interface for ADC7712
 * 	edit : 2009 / 07 / 08 By Ace version : 1.1
 *      edit : 2009 / 07 / 13 By Ace version : 1.2
 *      edit : 2009 / 08 / 4  By Ace version : 1.3
 */



#include <linux/config.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/mc146818rtc.h>
#include <linux/netdevice.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/delay.h>
#include <linux/devfs_fs_kernel.h>

//#define CONFIG_ADC_COMPARE 1 //only for invendis
#define NR_ADC_DEVICES 4
static DEFINE_SPINLOCK( ad7712_lock);
#define AD7712_MAJOR 233
#define CODE(hz)  ((10000000/512/hz )+ 0x0)
static int  ad7712_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg);

unsigned int ad7712_set[NR_ADC_DEVICES];
unsigned int tmp_cache;
static unsigned long  ad7712_read_status(void);
int Bits;


struct adc_state {
	//int  hz;
	int polar;
	char type;
	unsigned int range;
	int delay;
};

struct adc_offset {
	unsigned int range;
	char type;
	int  zero;
	int  full;
};

struct adc_state adc_states[NR_ADC_DEVICES];

#define ADC_SET		0xa001
#define ADC_GET		0xa002
#define ADC_GET_OFFSET  0xa003
#define ADC_SET_OFFSET  0xa004

#define BIPOLAR		0x0000
#define UNIPOLAR	0x1000

#define VOL		0
#define AMP		1

#define ADC_10V		0x000000
#define ADC_5V		0x040000
#define ADC_1V		0x0c0000
#define ADC_500mV	0x100000
#define ADC_150mV	0x180000

int Offset_VAL[5][3] = 
{{0x1BBD8,0x24E4D,0x7fffff}, //ok 10v
{0x1C6D6,0x22D15,0x7fffff}, //ok 5v
{0x22E86,0x24800,0x7374B1},//0x66685d}, //ok 1v
{0x21E5F,0x22000,0x7374B1}, //ok 500mv
{0x3B9C9,0x71900,0x7F991C}}; // 150mv

int Offset_AML[4][3] = 
{{0x1B0EE,0x27f00,0x5F8C51},
{0x1B0EE,0x27f00,0x5F8C51},
{0x1B0EE,0x27f00,0x5F8C51},
{0x1B0EE,0x27f00,0x5F8C51}};//0x3d6f77}};

unsigned long vol_divisor(unsigned int adc , int *div1 
				, int *div2 , int* rec1 
				, int *rec2 , int type , int num) {
	unsigned long vol = 0 , g = 10000;
	*div2 = 1;
	
	if(type) {
		*div2 = 100;
		*div1 = 2;
		g = g* 10;
		*rec1 = Offset_AML[num][0];
		*rec2 = Offset_AML[num][1];
		
	}else
	if((adc & 0x1c0000) == ADC_10V)	 {
		*div1 = 1;
		*rec1 = Offset_VAL[0][0];
		*rec2 = Offset_VAL[0][1];
		
	}else
	if((adc & 0x1c0000) == ADC_5V) {
		*div1 = 2;
		*rec1 = Offset_VAL[1][0];
		*rec2 = Offset_VAL[1][1];
		
	}
	else 
	if((adc & 0x1c0000) == ADC_1V){
		*div1 = 8;
		*rec1 = Offset_VAL[2][0];
		*rec2 = Offset_VAL[2][1];
		
	}
	else
	if((adc & 0x1c0000) == ADC_500mV) {
		*div1 = 16;
		*div2 = 100;
		g = g* 10;
		*rec1 = Offset_VAL[3][0];
		*rec2 = Offset_VAL[3][1];
		
	}
	else
	if((adc & 0x1c0000) == ADC_150mV) {
		*div1 = 66;
		*div2 = 100;
		g = g* 10;
		*rec1 = Offset_VAL[4][0];
		*rec2 = Offset_VAL[4][1];
	
	}
	
	vol = (0x7fffff*(*div1) /g);
	return vol;
}

int select_range(unsigned int adc,int type) {
	if(type)
		return 5;
	switch(adc & 0x1c0000) {
		case ADC_10V: 		return 0;
		case ADC_5V: 		return 1;
		case ADC_1V:     	return 2;
		case ADC_500mV:  	return 3;
		case ADC_150mV:  	return 4;
	}
	return 0;
}

int select_port(int i) {
	switch(i) {
		case 0:
			at91_set_gpio_output(AT91_PIN_PC1, 0);
			at91_set_gpio_output(AT91_PIN_PC2, 0);
			return 0;
		case 1:
			at91_set_gpio_output(AT91_PIN_PC1, 1);
			at91_set_gpio_output(AT91_PIN_PC2, 0);
			return 0;
		case 2:
			at91_set_gpio_output(AT91_PIN_PC1, 0);
			at91_set_gpio_output(AT91_PIN_PC2, 1);
			return 0;
		case 3:
			at91_set_gpio_output(AT91_PIN_PC1, 1);
			at91_set_gpio_output(AT91_PIN_PC2, 1);
			return 0;
		default:
			return -EINVAL;
	}
	return -EINVAL;
}

static void  ad7712_write_inst(u32 cmd)
{

	int i ;
	unsigned long control = cmd;
	
	at91_set_gpio_output(AT91_PIN_PD13, 1);
	at91_set_gpio_output(AT91_PIN_PD14, 0);
	udelay(100);
	at91_set_gpio_output(AT91_PIN_PB17, 1);
	at91_set_gpio_output(AT91_PIN_PB12, 1);
	
	at91_set_gpio_output(AT91_PIN_PB13, 0);
	at91_set_gpio_output(AT91_PIN_PB14, 0);
	at91_set_gpio_output(AT91_PIN_PC3, 0);
	at91_set_gpio_output(AT91_PIN_PB12, 0);
	//udelay(100); 
	udelay(5);
	for(i=0;i<24;i++)
	{
		 
		at91_set_gpio_output(AT91_PIN_PB13, 1);
		at91_set_gpio_output(AT91_PIN_PB16, 1);
		udelay(2);
		if (control & 0x00800000) {
			at91_set_gpio_output(AT91_PIN_PB14, 1);
		}
		else {
			at91_set_gpio_output(AT91_PIN_PB14, 0);
		}
	
		udelay(1);
		
		at91_set_gpio_output(AT91_PIN_PB13, 0);
		at91_set_gpio_output(AT91_PIN_PB16, 0);

		control<<=1;
		udelay(2);
	} 
	at91_set_gpio_output(AT91_PIN_PB13, 0); 
	at91_set_gpio_output(AT91_PIN_PB12, 1);
	at91_set_gpio_output(AT91_PIN_PB14, 0);
}
/*
		PB12 TF 
		PB13 TK
		PB14 TD
		PB15 RD 
		PB16 RK
		PB17 RF
		PC3  A0
		PB29 DRDY
*/
static unsigned long  ad7712_read_status(void)
{
	int ret ;//, i;
	unsigned long data = 0x00000000;
	at91_set_gpio_output(AT91_PIN_PD13, 0);
	at91_set_gpio_output(AT91_PIN_PD14, 1);
	udelay(100);
	at91_set_gpio_input(AT91_PIN_PB29, 0);
	
	at91_set_gpio_output(AT91_PIN_PB17, 1);
	at91_set_gpio_output(AT91_PIN_PB12, 1);

	at91_set_gpio_output(AT91_PIN_PB16, 0);
	at91_set_gpio_input(AT91_PIN_PB15, 0);
	at91_set_gpio_output(AT91_PIN_PB13, 0);
	at91_set_gpio_output(AT91_PIN_PC3, 0);
	
	at91_set_gpio_output(AT91_PIN_PB17, 0);
	udelay(5); 

	for(ret=0;ret<24;ret++)
	{
		data<<=1;

		at91_set_gpio_output(AT91_PIN_PB13, 1);
		at91_set_gpio_output(AT91_PIN_PB16, 1);
		udelay(2);
		if(at91_get_gpio_value(AT91_PIN_PB15)) {	
			data++;
		}
		udelay(1); 

		at91_set_gpio_output(AT91_PIN_PB13, 0);
		at91_set_gpio_output(AT91_PIN_PB16, 0);

		udelay(2); 
	}

	at91_set_gpio_output(AT91_PIN_PB16, 0);
	at91_set_gpio_output(AT91_PIN_PB17, 1);
	
	at91_set_gpio_input(AT91_PIN_PB15, 0);
	at91_set_gpio_output(AT91_PIN_PC3, 1);

	return data;
}


static unsigned long ad7712_read_data(void)
{
	int ret ,i;
	unsigned long data = 0x00000000;
	at91_set_gpio_output(AT91_PIN_PD13, 0);
	at91_set_gpio_output(AT91_PIN_PD14, 1);
	udelay(1000);
	at91_set_gpio_input(AT91_PIN_PB29, 0);
	
	at91_set_gpio_output(AT91_PIN_PB17, 1);
	at91_set_gpio_output(AT91_PIN_PB12, 1);
	
	for(i = 0 ;at91_get_gpio_value(AT91_PIN_PB29);i++) {
		udelay(100);
		if(i > 10000) {
				printk(KERN_WARNING "at91-ad7712 : Can't Get Data Ready\n");
				return 0xf000000;
			}
	}
	
	at91_set_gpio_output(AT91_PIN_PB16, 0);
	at91_set_gpio_input(AT91_PIN_PB15, 0);
	at91_set_gpio_output(AT91_PIN_PB13, 0);
	at91_set_gpio_output(AT91_PIN_PC3, 1);
	
	at91_set_gpio_output(AT91_PIN_PB17, 0);
	udelay(5); 

	for(ret=0;ret < 24;ret++)
	{
		data<<=1;

		at91_set_gpio_output(AT91_PIN_PB13, 1);
		at91_set_gpio_output(AT91_PIN_PB16, 1);
		udelay(2);
		if(at91_get_gpio_value(AT91_PIN_PB15)) {	
			data++;
		}
		udelay(1); 

		at91_set_gpio_output(AT91_PIN_PB13, 0);
		at91_set_gpio_output(AT91_PIN_PB16, 0);

		udelay(2); 
	}

	at91_set_gpio_output(AT91_PIN_PB16, 0);
	at91_set_gpio_output(AT91_PIN_PB17, 1);
	
	at91_set_gpio_input(AT91_PIN_PB15, 0);
	at91_set_gpio_output(AT91_PIN_PC3, 1);

	return data;

}

unsigned tmp_data[4] = {0,0,0,0};

int diff_data(int num , unsigned int data , unsigned int test) {
	unsigned int reg;
	if(test > 2)
		return 1;
	
	if(tmp_data[num] == 0) {
		tmp_data[num] = data;
		return 0;
	}else {
		reg = (tmp_data[num] > data ?
			tmp_data[num] - data :data-tmp_data[num]) * 100 / 0x7fffff; 
		if(reg > 2) {
			tmp_data[num] = data;
			return 1;
		}
	}
	tmp_data[num] = data;	
	return 0;
}

unsigned int ad7712_get_data(unsigned int ad7712_device) {
	unsigned int reg , data , data2 , test , tol = 0;
	int i = 0;
	select_port(ad7712_device);
	if(tmp_cache != ad7712_set[ad7712_device]) {
		do {
			ad7712_write_inst(ad7712_set[ad7712_device]);
			mdelay(1);
			reg = ad7712_read_status();
			i++;
			if(i > 1000) {
				printk("at91-ad7712 : Can't Write Data\n");
				return -1;
			}
			udelay(100);
		} while(reg != ad7712_set[ad7712_device]);
		tmp_cache = ad7712_set[ad7712_device];
		mdelay(30);
	}
	//User delay Time
	i = 0;
	mdelay(adc_states[ad7712_device].delay);
	
	if(ad7712_read_data() == 0xf000000) //first don't need 
		return 0xf000000;
#ifdef CONFIG_ADC_COMPARE
	do {
		mdelay(16);//60 Hz
		data2 = ad7712_read_data();
		tol += data2;
		if(data2 == 0xf000000)
			return 0xf000000;
		
		mdelay(16);//60 Hz
		data = ad7712_read_data();
		tol += data;
		if(data == 0xf000000)
			return 0xf000000;
		
		test = (data2 > data ? data2 - data : data - data2) * 100 / 0x7fffff;
		
		if(i > 10) {
			return tol / 20;
		}
		i++;
		
	}while(diff_data(ad7712_device,data , test)); // 2%
#else	
	mdelay(16);//60 Hz
	data = ad7712_read_data();
	if(data == 0xf000000)
		return 0xf000000;
#endif
	return data;
}

int offset_b;
unsigned int debug;
static int  ad7712_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{

	int ad7712_device = MINOR(inode->i_rdev);
	int rang , i;
	unsigned int data , data2 , test , tol = 0;
	struct adc_offset adcs;
	if (ad7712_device >= NR_ADC_DEVICES)
		return -ENODEV;


	
	switch (cmd) {
	case 0xa001:

		if (copy_from_user(&adc_states[ad7712_device], (struct adc_state *) arg,
			sizeof(struct adc_state)))
			return -EFAULT;

		ad7712_set[ad7712_device] = 0;
		if(adc_states[ad7712_device].type) {
			ad7712_set[ad7712_device] = 
				(0x28000+
				adc_states[ad7712_device].polar+
				CODE(60)+
				ADC_5V);
		}else{
			ad7712_set[ad7712_device] = 
				(0x28000+
				adc_states[ad7712_device].polar+
				CODE(60)+
				adc_states[ad7712_device].range);
		}
		//printk("%x\n",ad7712_set[ad7712_device]);

		break;
	case 0xa002:	
		return copy_to_user((struct adc_state *)arg, &adc_states[ad7712_device] ,sizeof(struct adc_state) );
	case 0xa003:
		if(ad7712_set[ad7712_device] == 0 || adc_states[ad7712_device].polar ) 
			return -EFAULT;
		
		rang = select_range(adc_states[ad7712_device].range,adc_states[ad7712_device].type);
		data = ad7712_get_data(ad7712_device);
		i = 0;
		#if 1
		do {
			
			data2 = ad7712_get_data(ad7712_device);
			if(data2 == 0xf000000)
				return -1;
			
			tol += data2;
			
			data = ad7712_get_data(ad7712_device);
			if(data == 0xf000000)
				return -1;
			
			tol += data;
			
			test = (data2 > data ? data2 - data : data - data2) * 100 / 0x7fffff;
			
			if(i > 10) {
				return tol / 20;
			}
			i++;
		
		}while(test > 2); // 2%
		#endif
		
		if(((int __user ) arg) == 0) {
			if(rang == 5)
				offset_b = (data < 0x7fffff ? 0x7fffff - data : Offset_AML[ad7712_device][0]);
			else
				offset_b = (data < 0x7fffff ? 0x7fffff - data : Offset_VAL[rang][0]);
			return offset_b;
		}else {
			if(offset_b) {
				data = data+offset_b;
			}
			else {
				if(rang == 5)
					data = data+Offset_AML[ad7712_device][0];
				else
					data = data+Offset_VAL[rang][0];
			}
			offset_b = 0;
			if(rang == 5)
				return (Offset_AML[ad7712_device][2]*2) > data ? 
				(Offset_AML[ad7712_device][2]*2) - data : Offset_AML[ad7712_device][1];
			else
				return (Offset_VAL[rang][2]*2) > data ? 
				(Offset_VAL[rang][2]*2) - data : Offset_VAL[rang][1];
		}
		break;
	case 0xa004:
		
		if (copy_from_user(&adcs, (struct adc_offset *) arg,
		    sizeof(struct adc_offset)))
			return -EFAULT;
		if(adcs.type == 1) {
			Offset_AML[ad7712_device][0] =
					adcs.zero;
			Offset_AML[ad7712_device][1] =
					adcs.full;
		}else
		{
			Offset_VAL[select_range(adcs.range,adcs.type)][0] =
					adcs.zero;
			Offset_VAL[select_range(adcs.range,adcs.type)][1] =
					adcs.full;
		}
		return 0;
		break;
	case 0xa005:
		*((unsigned int __user *) arg) = debug;
		return 0;
		break;
	default:
		
		return -EINVAL;

	}
	return 0;

}

static int  ad7712_open(struct inode *inode, struct file *file)
{
	unsigned int ad7712_device = MINOR(inode->i_rdev);

	if (ad7712_device >= NR_ADC_DEVICES)
		return -ENODEV;

	file->private_data = (void *)ad7712_device;
	
	return 0;

}

static int ad7712_close(struct inode *inode, struct file *file)
{
	
	return 0;
}

/*
 *	The various file operations we support.
 */
static ssize_t ad7712_read(struct file *file, char *buf, size_t count, loff_t *offset)
{
	
	spin_lock(&ad7712_lock);
	unsigned int ad7712_device = (unsigned int) file->private_data;
	unsigned int data;//reg
	unsigned long res1 , res2 ;
	long res3;
	unsigned char bufs[128];
	char flag = 1;
	int div1 , div2 , rec1 , rec2;
	
	data = ad7712_get_data(ad7712_device);
	if(data == 0xf000000) {
			copy_to_user (buf, "-1", 2);
			return 0;//-ENOMEM;
	}
	debug = data;
	res1 =  vol_divisor(ad7712_set[ad7712_device] , &div1 ,&div2,&rec1 ,&rec2 
			, adc_states[ad7712_device].type ,ad7712_device);
	
	data = data + rec1;
	
	//printk("data = %x\n",data);
	if(!adc_states[ad7712_device].polar) {
		flag = (data & 0x800000?1:0);
		data =(data & 0x800000?data - 0x7fffff : 0x7fffff - data);
	}
	data = data + ((rec2 * (data/8388))/1000);
	
	//if(!flag &&  adc_states[ad7712_device].range == ADC_150mV )
	//	data = (data > 0x2eea2) ?data - 0x2eea2:data ;
	//if(!flag &&  adc_states[ad7712_device].range == ADC_500mV )
	//	data = (data > 0xbfff) ?data - 0xbfff:data ;
	//printk("data = %x  rec2 = %x\n",data,rec2);
	//printk("data = %x   %x\n",data,((rec2 * (data/8388))/1000));
	res3 = (((data)*div2) / res1);
	if(adc_states[ad7712_device].type) { 
		//printk("VOL : %ld  , new : %ld\n",res3,(res3/120));
		res3 = (res3 /120) - ((90 *res3 /120)/20000);
	}
	res2 = ((res3/1000)?res3-((res3/1000)*1000):res3);
	//memset(bufs , '\0' , 128);
	sprintf(bufs,"%c%ld.%s%ld"
			,flag?'+':'-',res3/1000,((res2 < 10) ?"00":(res2 < 100)?"0":""),
			res2);
	
	copy_to_user (buf, bufs, sizeof(bufs));
	//printk("%s\n",bufs);
	spin_unlock(&ad7712_lock);
	return strlen(bufs);
}
static struct file_operations  ad7712_fops = {
	.read    =  ad7712_read,
	.ioctl   =  ad7712_ioctl,
	.open    =  ad7712_open,
	.release =  ad7712_close,
};

/*static struct miscdevice  ad7712_dev = {
	MISC_AD7712_MINOR,
	"adc",
	&ad7712_fops
};*/

static int  ad7712_init(void)
{
	int i;
	//unsigned long data;

	//pr_info("Artila AD7712 driver Version 1.0\n");
	if (register_chrdev(AD7712_MAJOR, "adc", &ad7712_fops)) {
		printk(KERN_ERR "at91_ad7712: Unable to get major %d for AD7712\n"
		, AD7712_MAJOR);
		return -EIO;
	}

	for (i = 0; i < NR_ADC_DEVICES; i++) {
		devfs_mk_cdev(MKDEV(AD7712_MAJOR, i), S_IFCHR | S_IRUSR | S_IWUSR, "adc/%d",i);
		ad7712_set[i] = 0;
	}

	offset_b = 0;
	for(i = 0 ;i < 4 ; i++)
		adc_states[i].delay = 100;
	#ifdef CONFIG_ADC_COMPARE
	pr_info("Artila AD7712 driver Version 1.3.1 (Invendis) \n");
	#else
	pr_info("Artila AD7712 driver Version 1.3.1\n");
	#endif
	tmp_cache = 0;
	debug = 0;
	return 0;		
	
}

static void __exit  ad7712_exit(void)
{
	int i;
	for (i = 0; i < NR_ADC_DEVICES; i++) {
		devfs_remove("adc/%d", i);
	}
	
	if (unregister_chrdev(AD7712_MAJOR, "adc")) {
		printk(KERN_ERR "at91_ad7712: Unable to release major %d for AD7712\n", AD7712_MAJOR);
		return;
	}
	//misc_deregister(& ad7712_dev);
}



module_init( ad7712_init);
module_exit( ad7712_exit);

MODULE_AUTHOR("Ace Yang");
MODULE_LICENSE("GPL");
