/*
 * GPIO driver for Artila Matrix-500 based on Atmel AT91RM9200 (Thunder)
 *
 *  Copyright (C) 2005 Artila Electronics,Ltd
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <asm/bitops.h> 
#include <asm/uaccess.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <asm/arch/gpio.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>
#include <asm/arch/AT91RM9200_EMAC.h>
#include <linux/mii.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/delay.h> 
#include <linux/proc_fs.h>	/* Necessary because we use the proc fs Ace*/
//#include <linux/signal.h>

#define GPIO_MINOR 		254

#define procfs_name "gpio"
struct proc_dir_entry *matrix_proc_file;



#if defined(CONFIG_MACH_M501)
#define MATRIX500_GPIO_COUNT	24 //for matrix520, 24 //for m501, 16 for m500
#define DIP_SW_COUNT		2 //for matrix520, 2:for m501
#define DIP_SW1			MATRIX500_GPIO_COUNT
#define DIP_SW2			(MATRIX500_GPIO_COUNT+1)
static int MATRIX500_GPIO_MAP[MATRIX500_GPIO_COUNT+DIP_SW_COUNT][3]={
	{AT91_PIN_PD24,3,0x01000000} ,
	{AT91_PIN_PA19,0,0x00080000} ,
	{AT91_PIN_PD25,3,0x02000000},
	{AT91_PIN_PB2 ,1,0x00000004},
	{AT91_PIN_PB8 ,1,0x00000100},
	{AT91_PIN_PB9 ,1,0x00000200},
	{AT91_PIN_PB10,1,0x00000400},
	{AT91_PIN_PB11,1,0x00000800},
	{AT91_PIN_PC0 ,2,0x00000001},
	{AT91_PIN_PC1 ,2,0x00000002},
	{AT91_PIN_PC2 ,2,0x00000004},
	{AT91_PIN_PC3 ,2,0x00000008},
	{AT91_PIN_PC5 ,2,0x00000020},
	//AT91_PIN_PC6,
	{AT91_PIN_PB22,1,0x00400000},
	{AT91_PIN_PB28,1,0x10000000},
	{AT91_PIN_PB29,1,0x20000000},
	{AT91_PIN_PC14,2,0x00004000},
	{AT91_PIN_PC15,2,0x00008000},
	{AT91_PIN_PD7 ,3,0x00000080},
	{AT91_PIN_PD8 ,3,0x00000100},
	{AT91_PIN_PD9 ,3,0x00000200},
	{AT91_PIN_PD10,3,0x00000400},
	{AT91_PIN_PD11,3,0x00000800},
	{AT91_PIN_PD12,3,0x00001000},
	//Switch
	{AT91_PIN_PD23,3,0x00800000},
	{AT91_PIN_PD22,3,0x00400000}
	/*AT91_PIN_PD16,
	AT91_PIN_PD18,
	AT91_PIN_PD27*/
};

#elif defined(CONFIG_MACH_M501_SOM)
#define MATRIX500_GPIO_COUNT	32 //for matrix520, 24 //for m501, 16 for m500
#define DIP_SW_COUNT		2 //for matrix520, 2:for m501
#define DIP_SW1			MATRIX500_GPIO_COUNT
#define DIP_SW2			(MATRIX500_GPIO_COUNT+1)
static int MATRIX500_GPIO_MAP[MATRIX500_GPIO_COUNT+DIP_SW_COUNT][3]={
	{AT91_PIN_PD24,3,0x01000000} ,
	{AT91_PIN_PA19,0,0x00080000} ,
	{AT91_PIN_PD25,3,0x02000000},
	{AT91_PIN_PB2 ,1,0x00000004},
	{AT91_PIN_PB8 ,1,0x00000100},
	{AT91_PIN_PB9 ,1,0x00000200},
	{AT91_PIN_PB10,1,0x00000400},
	{AT91_PIN_PB11,1,0x00000800},
	{AT91_PIN_PC0 ,2,0x00000001},
	{AT91_PIN_PC1 ,2,0x00000002},
	{AT91_PIN_PC2 ,2,0x00000004},
	{AT91_PIN_PC3 ,2,0x00000008},
	{AT91_PIN_PC5 ,2,0x00000020},
	//AT91_PIN_PC6,
	{AT91_PIN_PB22,1,0x00400000},
	{AT91_PIN_PB28,1,0x10000000},
	{AT91_PIN_PB29,1,0x20000000},
	{AT91_PIN_PC14,2,0x00004000},
	{AT91_PIN_PC15,2,0x00008000},
	{AT91_PIN_PD7 ,3,0x00000080},
	{AT91_PIN_PD8 ,3,0x00000100},
	{AT91_PIN_PD9 ,3,0x00000200},
	{AT91_PIN_PD10,3,0x00000400},
	{AT91_PIN_PD11,3,0x00000800},
	{AT91_PIN_PD12,3,0x00001000},
	{AT91_PIN_PD13,3,0x00002000},
	{AT91_PIN_PD14,3,0x00004000},
	{AT91_PIN_PD15,3,0x00008000},
	{AT91_PIN_PD16,3,0x00010000},
	{AT91_PIN_PD17,3,0x00020000},
	{AT91_PIN_PD18,3,0x00040000},
	{AT91_PIN_PD26,3,0x04000000},
	{AT91_PIN_PD27,3,0x08000000},
	{AT91_PIN_PD23,3,0x00800000},
	{AT91_PIN_PD22,3,0x00400000}
};
#endif

#define MATRIX500_TOTAL_GPIO	(MATRIX500_GPIO_COUNT+DIP_SW_COUNT)
/* IOCTL Commands */
#define GPIO_IOCTL_COUNT	0
#define GPIO_IOCTL_DIPSW	1
#define GPIO_IOCTL_GET		2
#define	GPIO_IOCTL_SET		3
#define	GPIO_IOCTL_CLEAR	4
#define GPIO_IOCTL_OUTPUT	5
#define GPIO_IOCTL_INPUT	6
#define GPIO_IOCTL_MODE		7
//#define GPIO_IOCTL_INPUT_SET	8
#define GPIO_IOCTL_STATUS	0xa001

#define GPIO_INTERRUPTS_EN   0xb001
#define GPIO_INTERRUPTS_DS   0xb002
#define GPIO_INTERRUPTS_CR   0xb003
#define GPIO_INTERRUPTS_TS   0xb004

//#define MY_WORK_QUEUE_NAME "WQsched.c"

//static struct workqueue_struct *my_workqueue;


static DEFINE_SPINLOCK(gpio_lock);
static int gpio_open_cnt=0;
static int matrix500_gpio_mode[MATRIX500_TOTAL_GPIO];

unsigned int gpio_irq = 0;
int gpio_i = 0;


struct gpio_irqs {
        int trigger[MATRIX500_TOTAL_GPIO];
        int num[MATRIX500_TOTAL_GPIO];
	long MATRIX500_IRQ_MAP[MATRIX500_TOTAL_GPIO];
};
struct gpio_irqs gpio_irqs;

struct gpio_set_irq {
        int trigger;
        int pin;
};


/* ......................................................................... */


/* ......................................................................... */


irqreturn_t irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	/* 
	 * This variables are static because they need to be
	 * accessible (through pointers) to the bottom half routine.
	 */
	int i;
	unsigned int status[4][2];

	status[0][0] = AT91_SYS->PIOA_ISR & 0x0008001f;
	status[0][1] = AT91_SYS->PIOA_PDSR;
	status[1][0] = AT91_SYS->PIOB_ISR ;
	status[1][1] = AT91_SYS->PIOB_PDSR;
	status[2][0] = AT91_SYS->PIOC_ISR & 0x0000fd7f;
	status[2][1] = AT91_SYS->PIOC_PDSR;
	status[3][0] = AT91_SYS->PIOD_ISR;
	status[3][1] = AT91_SYS->PIOD_PDSR;

	if(status[0][0] | status[1][0] | status[2][0] | status[3][0]) {

		for(i = 0 ; i < MATRIX500_TOTAL_GPIO ; i++) {
			if(gpio_irqs.MATRIX500_IRQ_MAP[i] >= 0 && gpio_irqs.num[i]) {

			if((status[MATRIX500_GPIO_MAP[i][1]][0] & MATRIX500_GPIO_MAP[i][2]) 
				&& !(gpio_irqs.trigger[i] ^ 
				((status[MATRIX500_GPIO_MAP[i][1]][1] & MATRIX500_GPIO_MAP[i][2])
				?1:0))) {
					gpio_irqs.MATRIX500_IRQ_MAP[i]++;
				
				}
		}
		}
	}


	return IRQ_HANDLED;
}
/*
 * GPIO device is opened, and GPIO starts running.
 */
static int matrix500_gpio_open(struct inode *inode, struct file *file)
{
	spin_lock(&gpio_lock);
	gpio_open_cnt++;
	spin_unlock(&gpio_lock);
	return 0;
}

/*
 * Close the GPIO device.
 * If CONFIG_GPIO_NOWAYOUT is NOT defined then the GPIO is also
 *  disabled.
 */
static int matrix500_gpio_close(struct inode *inode, struct file *file)
{
	spin_lock(&gpio_lock);
	spin_unlock(&gpio_lock);
	return 0;
}

int
procfile_read(char *page,char **start,off_t offset, int count, int *eof, void *data)
{
	int len, i , j = 0 , k = 0 ;
	unsigned char buf[4096];
	//k = sprintf(&buf[k], "Gpio %s Trigger :\n",gpio_i?"High":"Low");
	for(i = 0 ; i < MATRIX500_TOTAL_GPIO ; i++) {
		if(gpio_irqs.num[i] == 1 ) {
			j = sprintf(&buf[k], "Gpio%d : %s Trigger , count:%d\n",i
				,gpio_irqs.trigger[i]?"High":"Low", gpio_irqs.MATRIX500_IRQ_MAP[i]);
			k+=j;
		}
	}
	if(k == 0)
		len = sprintf(page,"%s","(NULL)\n");
	else 
		len = sprintf(page, "%s",buf);
		
	return len;
}

/*
 * Handle commands from user-space.
 */
static int matrix500_gpio_ioctl(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	int i;
	unsigned long status = 0x00000000;
	
	switch(cmd){
		case GPIO_IOCTL_COUNT:
			*((int __user *) argp) = MATRIX500_GPIO_COUNT; 
		return 0;

		case GPIO_IOCTL_DIPSW:
			*((int __user *) argp) = DIP_SW_COUNT; 
		return 0;

		case GPIO_IOCTL_GET:
			i = *((int __user *) argp);
			if(i >= MATRIX500_TOTAL_GPIO)
				return -EINVAL;	
		return at91_get_gpio_value(MATRIX500_GPIO_MAP[i][0]);

		case GPIO_IOCTL_SET:
			i = *((int __user *) argp);
			if(i >= MATRIX500_GPIO_COUNT  )
				return -EINVAL;	
			if(matrix500_gpio_mode[i] == 1)
				at91_set_gpio_input(MATRIX500_GPIO_MAP[i][0], 1);
			else
				at91_set_gpio_value(MATRIX500_GPIO_MAP[i][0], 1);
		return 0;

		case GPIO_IOCTL_CLEAR:
			i = *((int __user *) argp);
			if(i >= MATRIX500_GPIO_COUNT  )
				return -EINVAL;	
			if(matrix500_gpio_mode[i] == 1)
				at91_set_gpio_input(MATRIX500_GPIO_MAP[i][0], 0);
			else
				at91_set_gpio_value(MATRIX500_GPIO_MAP[i][0], 0);
		return 0;
		
		case GPIO_IOCTL_OUTPUT:

			i = *((int __user *) argp);
			if(i >= MATRIX500_GPIO_COUNT )
				return -EINVAL;
			matrix500_gpio_mode[i] = 0;
			at91_set_gpio_output(MATRIX500_GPIO_MAP[i][0],0);

			gpio_irqs.MATRIX500_IRQ_MAP[i] = -1;
			gpio_irqs.num[i] = 0;
			gpio_irqs.trigger[i] = -1;

			return 0;
			

		case GPIO_IOCTL_INPUT:

			i = *((int __user *) argp);
			if(i >= MATRIX500_GPIO_COUNT)
				return -EINVAL;
			matrix500_gpio_mode[i] = 1;
			at91_set_gpio_input(MATRIX500_GPIO_MAP[i][0], 0);
			
			return 0;
			
		case	GPIO_IOCTL_MODE:
			i = *((int __user *) argp);
			if(i >= MATRIX500_GPIO_COUNT)
				return -EINVAL;
			return matrix500_gpio_mode[i];

	
		/*INTERRUPTS for Artila Ace*/
		case	GPIO_INTERRUPTS_EN: {

				struct gpio_set_irq set_irq;
	
				if (copy_from_user
				(&set_irq, (struct  gpio_set_irq*) arg,
				sizeof(struct gpio_set_irq)))
					return -EFAULT;
	
				if(set_irq.pin >= MATRIX500_TOTAL_GPIO)
					return -EINVAL;
	
				gpio_irqs.num[set_irq.pin] = 1;
				gpio_irqs.trigger[set_irq.pin] = set_irq.trigger; 
				gpio_irqs.MATRIX500_IRQ_MAP[set_irq.pin] = 0;
				
				if(!gpio_irq) {
					
					matrix_proc_file = create_proc_entry(procfs_name, 0644, NULL);
					if (matrix_proc_file == NULL) 
					{
						remove_proc_entry(procfs_name, &proc_root);
						printk(KERN_ALERT "Error: Could not initialize /proc/%s\n",
						procfs_name);
						return -ENOMEM;
					}
					matrix_proc_file->read_proc = procfile_read;
					matrix_proc_file->owner 	 = THIS_MODULE;
					matrix_proc_file->mode 	 = S_IFREG | S_IRUGO;
					
					if(request_irq(1,irq_handler,SA_SHIRQ, "matrix_gpio",irq_handler))
							return -EBUSY;//Ace
					//printk("Matrix GPIO Interrupt Enable , %s Trigger.\n",gpio_i?"High":"Low");
					gpio_irq = 1;
				}
			

			return 0;
			}
		case 	GPIO_INTERRUPTS_DS:
			
			i = *((int __user *) argp);
			if(i >= MATRIX500_TOTAL_GPIO || gpio_irqs.num[i] == 0)
				return -EINVAL;
			
			gpio_irqs.num[i] = 0;
			gpio_irqs.trigger[i] = -1;
			gpio_irqs.MATRIX500_IRQ_MAP[i] = -1;
			
			//printk("Matrix GPIO Interrupt Disable.\n");
			return 0;
		case	GPIO_INTERRUPTS_CR:
			i = *((int __user *) argp);
			if(matrix500_gpio_mode[i] && i < MATRIX500_TOTAL_GPIO && gpio_irq && gpio_irqs.num[i] == 1) {
				gpio_irqs.MATRIX500_IRQ_MAP[i] = 0 ;
				return 0;
			}
			return -EINVAL;
		case	GPIO_INTERRUPTS_TS:
			i = *((int __user *) argp);
			if((i >= MATRIX500_TOTAL_GPIO) || !matrix500_gpio_mode[i] || !gpio_irq || gpio_irqs.num[i] != 1)
				return -EINVAL;
			return gpio_irqs.MATRIX500_IRQ_MAP[i];

		/*Power Management for Artila*/
		case 	9:// Ace
			powerdown_phy();	
			mdelay(1);
			at91_set_gpio_output(AT91_PIN_PD6, 1);
			AT91_SYS->PMC_SCDR = AT91C_PMC_UDP| AT91C_PMC_MCKUDP |AT91C_PMC_UHP | AT91C_PMC_PCK;
			AT91_SYS->EBI_SDRC_LPR = AT91C_SDRC_LPCB;
			AT91_SYS->PMC_PCDR = 0xffffffc0;//Ace
			mdelay(10);
			return 0;
		case 	10 :	
			at91_set_gpio_output(AT91_PIN_PD6, 0);
			AT91_SYS->PMC_SCER = AT91C_PMC_UDP| AT91C_PMC_MCKUDP |AT91C_PMC_UHP | AT91C_PMC_PCK;
			AT91_SYS->EBI_SDRC_LPR = AT91C_SDRC_LPCB;
			AT91_SYS->PMC_PCER = 0xffffffff;//Ace
			mdelay(1);
			reset_phy();
			mdelay(10);
			return 0;
		case	0xa001:
			for (i = MATRIX500_TOTAL_GPIO ; i >= 0 ; i--) {
				status <<= 1;
				if(at91_get_gpio_value(MATRIX500_GPIO_MAP[i][0]))
					status++;
			}
			*((unsigned long __user *) argp) = status; 
			return 0;		
		default:return -ENOIOCTLCMD; 
	}
	return -ENOIOCTLCMD;
}

/*static ssize_t matrix500_gpio_read(struct file *file, char *data, size_t len, loff_t *ppos)
{
	spin_lock(&gpio_lock);
	spin_unlock(&gpio_lock);
	return len;
}*/

/*
 * Pat the GPIO whenever device is written to.
 */
/*static ssize_t matrix500_gpio_write(struct file *file, const char *data, size_t len, loff_t *ppos)
{
	spin_lock(&gpio_lock);
	spin_unlock(&gpio_lock);
	return len;
}*/

/* ......................................................................... */

static struct file_operations matrix500gpio_fops =
{
	.owner		= THIS_MODULE,
	.ioctl		= matrix500_gpio_ioctl,
	.open		= matrix500_gpio_open,
	.release	= matrix500_gpio_close,
	//.read		= matrix500_gpio_read,
	/*.write		= matrix500_gpio_write,*/
};

static struct miscdevice matrix500_gpio_miscdev =
{
	.minor		= GPIO_MINOR,
	.name		= "gpio",
	.fops		= &matrix500gpio_fops,	
};

static int __init matrix500_gpio_init(void)
{
	int res;
	int i;

	res = misc_register(&matrix500_gpio_miscdev);
	if (res)
		return res;
	
	for(i=0;i < MATRIX500_GPIO_COUNT ; i++) {
		matrix500_gpio_mode[i] = 1;
		at91_set_gpio_input(MATRIX500_GPIO_MAP[i][0], 0);
	}
	if(DIP_SW_COUNT != 0) {
		for(;i < MATRIX500_TOTAL_GPIO ; i++) {
			//matrix500_gpio_mode[i] = 1;
			at91_set_gpio_input(MATRIX500_GPIO_MAP[i][0], 1);
		}
	}
	//Ace
	memset (gpio_irqs.MATRIX500_IRQ_MAP , -1 ,MATRIX500_TOTAL_GPIO);
	memset (gpio_irqs.num , 0 ,MATRIX500_TOTAL_GPIO);
	memset (gpio_irqs.trigger , -1 ,MATRIX500_TOTAL_GPIO);

	printk("Matrix500 GPIO Driver Ver 1.06 Loaded.\n");
	return 0;
}

static void __exit matrix500_gpio_exit(void)
{
	misc_deregister(&matrix500_gpio_miscdev);
	
}

module_init(matrix500_gpio_init);
module_exit(matrix500_gpio_exit);

MODULE_LICENSE("GPL")
MODULE_AUTHOR("Alan Lu")
MODULE_DESCRIPTION("GPIO driver for Matrix500 based on Atmel AT91RM9200")
