/*
 * LCD interface for WG16080A
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1996, 1997 by Andrew Bose
 *
 * Linux kernel version history:
 *       March 2001: Ported from 2.0.34  by Liam Davies
 *
 */



#include <linux/config.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/mc146818rtc.h>
#include <linux/netdevice.h>
#include <linux/sched.h>
#include <linux/delay.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/delay.h>

#include "lcd_WG16080A.h"

//static DEFINE_SPINLOCK(lcd_lock);

static int lcd_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg);

static unsigned int lcd_present = 1;

/* used in arch/mips/cobalt/reset.c */
//int led_state = 0;
//alan
static int at91_lcd_base;
static u8 lcd_read_status(void);
int Mode;

#define BusyCheck()	
#define MISC_LCD_MINOR 240



static void lcd_write_inst(u8 cmd)
{
	udelay(100);
	u8 __iomem* pLCD = (void __iomem *)(at91_lcd_base);
	*(pLCD+0x01) = cmd;
}

static u8 lcd_read_status(void)
{
	udelay(200);
	u8 __iomem* pLCD = (void __iomem *)(at91_lcd_base);
	
	//return *(pLCD+0x04);
	return *(pLCD+0x5);
}



static void lcd_write_data(u8 data)
{
	udelay(100);
	u8 __iomem* pLCD = (void __iomem *)(at91_lcd_base);
	*(pLCD) = data;

}

static u8 lcd_read_data(void)
{
	udelay(200);
	u8 __iomem* pLCD = (void __iomem *)(at91_lcd_base);
	
	 //return *(pLCD+0x5);
	return *(pLCD+0x04);
}

static void _lcd_init_mode(u8 cmd) {
	lcd_write_inst(0x00);
	lcd_write_data(cmd);
}

static void lcd_init_mode(u8 cmd) {

	_lcd_init_mode(cmd);
	lcd_write_inst(0x01);
	lcd_write_data(0xf7);
	lcd_write_inst(0x02);
	lcd_write_data(0x13);
	lcd_write_inst(0x03);
	lcd_write_data(MaxRow_1);
	lcd_write_inst(0x0a);
	lcd_write_data(0x00);
	lcd_write_inst(0x0b);
	lcd_write_data(0x00);	
}

void graphic_index(int x , int y ){//, char** s) {
	unsigned long a ,b ;
	
	if((a=(y * 0x01) + (x *0x14)) >= 0x100) {
	//if((y > 17 && x >= 12) || x >= 13) {
		b=(((y * 0x01) + (x *0x14))/0x100);
		a=((y * 0x01) + (x *0x14)) - (b*0x100);
	}else{
		
		b=0x00;
	}
	
	lcd_write_inst(0x0a);
	lcd_write_data(a);
	lcd_write_inst(0x0b);
	lcd_write_data(b);

}

static int lcd_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	//unsigned long address, a;
	int i , j , index1 , index2;
	
	switch (cmd) {
	case 1:
	
		lcd_init_mode(0x3c);
		Mode = 0;
		break;
	
	case 2:
		
		lcd_init_mode(0x32);
		Mode = 1;
		break;
	case 3:
		if(Mode)
			_lcd_init_mode(0x12);
		else
			_lcd_init_mode(0x1c);
		lcd_write_inst(0x0a);
		lcd_write_data(0x00);
		lcd_write_inst(0x0b);
		lcd_write_data(0x00);
		
		for(i = (Mode?1600:160) ; i >= 0; i--) {
			lcd_write_inst(0x0c);
			lcd_write_data(0x00);
		}
		if(Mode)
			_lcd_init_mode(0x32);
		else
			_lcd_init_mode(0x3c);
		break;
	case 4: 
	{
		struct lcd_display display;
		if (copy_from_user(&display, (struct lcd_display *) arg,
		    sizeof(struct lcd_display)))
			return -EFAULT;
		index1 = strlen(display.buffer);
		graphic_index(display.x,display.y);
		for(i = 0 ; i < index1 ; i++) {
			lcd_write_inst(0x0c);
			lcd_write_data(display.buffer[i]);
		}
		break;
	}
	case 5:
	{
		struct lcd_display display;
		unsigned char p;
		if (copy_from_user(&display, (struct lcd_display *) arg,
			sizeof(struct lcd_display)))
			return -EFAULT;
		
		index1 = display.row-1 ;
		index2 = display.column / 8;
		//printk("x=%d , y=%d , row=%d , column=%d\n ",
		//       display.x,display.y,display.row,display.column);
		if(display.flag) {
			/*if(display.flag == 0xff) {
				for(i = index1 ; i >= 0 ; i--) {
					graphic_index(((index1 + display.x) - i),display.y);
					for( j = 0 ; j < index2 ; j++) {
						p = little_end(display.buffer[(i*index2)+j]);
						lcd_write_inst(0x0c);
						lcd_write_data(~p);
					}
				}
			}else*/ {
				for(i = index1 ; i >= 0 ; i--) {
					graphic_index(((index1 + display.x) - i),display.y);
					for( j = 0 ; j < index2 ; j++) {
						p = little_end(display.buffer[(i*index2)+j]);
						lcd_write_inst(0x0c);
						lcd_write_data(p);
					}
				}
			}	
		}else {
			for(i = 0 ; i <= index1 ; i++) {
				graphic_index((display.x + i),display.y);
				for( j = 0 ; j < index2 ; j++) {
					p = little_end(display.buffer[(i*index2)+j]);
					lcd_write_inst(0x0c);
					lcd_write_data(p);
				}
			}
		}
		
		break;
		case 6:
			struct lcd_display display;
			unsigned char p;
			int i;
			if (copy_from_user(&display, (struct lcd_display *) arg,
			    sizeof(struct lcd_display)))
				return -EFAULT;
		
			index1 = display.row-1 ;
			index2 = display.column / 8;
			graphic_index((display.x + i),display.y);
			lcd_write_inst(0x0d);
			lcd_write_data(p);
			printf("p = %x\n",p);
			i = display.column % 8;
			i*=0x1;
			p |= i;
			graphic_index((display.x + i),display.y);
			lcd_write_inst(0x0c);
			lcd_write_data(p);
		break;
		
	}




	default:
		
		return -EINVAL;

	}
	udelay(20);
	
	
	return 0;

}

static int lcd_open(struct inode *inode, struct file *file)
{
	if (!lcd_present)
		return -ENXIO;
	else
		return 0;
}



/*
 *	The various file operations we support.
 */

static struct file_operations lcd_fops = {
	//.read = lcd_read,
	.ioctl = lcd_ioctl,
	.open = lcd_open,
};

static struct miscdevice lcd_dev = {
	MISC_LCD_MINOR,
	"lcd",
	&lcd_fops
};

static int lcd_init(void)
{
	int ret;
	//unsigned long data;

	pr_info("%s\n", LCD_DRIVER);
	ret = misc_register(&lcd_dev);
	if (ret) {
		printk(KERN_WARNING "Unable to register misc device.\n");
		return ret;
	}
	/* map physical adress */
	at91_lcd_base = (unsigned long) ioremap(0x70000000, SZ_16M);
	if(!at91_lcd_base) {
		printk("ioremap AT91 LCD failed\n");
		return 0;
	}
	
	AT91_SYS->EBI_CSA &= ~AT91C_EBI_CS4A; 
	AT91_SYS->EBI_SMC2_CSR[6] = 0xff034fff;

	at91_set_gpio_output(AT91_PIN_PD24, 0);
	at91_set_gpio_output(AT91_PIN_PB18, 0);
	udelay(2000);
	at91_set_gpio_output(AT91_PIN_PB18, 1);
	return 0;		
	
}

static void __exit lcd_exit(void)
{
	misc_deregister(&lcd_dev);
}



module_init(lcd_init);
module_exit(lcd_exit);

MODULE_AUTHOR("Ace Yang");
MODULE_LICENSE("GPL");
