/*
 * GPIO driver for Artila Matrix-500 based on Atmel AT91RM9200 (Thunder)
 *
 *  Copyright (C) 2005 Artila Electronics,Ltd
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <asm/bitops.h> 
#include <asm/uaccess.h>
#include <linux/init.h>
#include <linux/spinlock.h>
#include <asm/arch/gpio.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>

#define GPIO_MINOR 		253


#define PM_E	0
#define PM_D	1
static DEFINE_SPINLOCK(pm_lock);
int gpio_open_cnt = 0;
/*
 * GPIO device is opened, and GPIO starts running.
 */
static int matrix_pm_open(struct inode *inode, struct file *file)
{
	spin_lock(&pm_lock);
	gpio_open_cnt++;
	spin_unlock(&pm_lock);
	return 0;
}

/*
 * Close the GPIO device.
 * If CONFIG_GPIO_NOWAYOUT is NOT defined then the GPIO is also
 *  disabled.
 */
static int matrix_pm_close(struct inode *inode, struct file *file)
{
	spin_lock(&pm_lock);
	spin_unlock(&pm_lock);
	return 0;
}

/*
 * Handle commands from user-space.
 */
static int matrix_pm_ioctl(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	int i;
	
	
	switch(cmd){
		case PM_E:
			
		return 0;

		default:return -ENOIOCTLCMD; 
	}
	return -ENOIOCTLCMD;
}

static struct file_operations matrixgpio_fops =
{
	.owner		= THIS_MODULE,
	.ioctl		= matrix_pm_ioctl,
	.open		= matrix_pm_open,
	.release	= matrix_pm_close,
	/*.read		= matrix_pm_read,
	.write		= matrix_pm_write,*/
};

static struct miscdevice matrix_pm_miscdev =
{
	.minor		= GPIO_MINOR,
	.name		= "mpm",
	.fops		= &matrixgpio_fops,	
};

static int __init matrix_pm_init(void)
{
	int res;
	int i;

	res = misc_register(&matrix_pm_miscdev);
	if (res)
		return res;
	
	printk("Matrix PM Driver Loaded.\n");
	return 0;
}

static void __exit matrix_pm_exit(void)
{
	misc_deregister(&matrix_pm_miscdev);

}

module_init(matrix_pm_init);
module_exit(matrix_pm_exit);

MODULE_LICENSE("GPL")
MODULE_AUTHOR("Ace Yang")
MODULE_DESCRIPTION("PM driver for Matrix500 based on Atmel AT91RM9200")
