#include <linux/config.h>
#include <sound/driver.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <asm/arch-at91rm9200/artila.h>
#ifdef CONFIG_PM
#include <linux/pm.h>
#endif

#include <asm/hardware.h>
#include <asm/arch/AT91RM9200_SSC.h>
#include <asm/mach-types.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>

#define TLV320_MAJOR 234

static int cur_left_vol;
static int cur_right_vol;

static struct i2c_client * tlv320_client = NULL;

int tlv320_set_volumes(uint left_vol, uint right_vol) {
        unsigned short voldata;

        if (!tlv320_client)
                return -1;

	if (left_vol  > 100) left_vol  = 100;
	if (right_vol > 100) right_vol = 100;

        cur_left_vol = left_vol;
        cur_right_vol = right_vol;

	//if(left_vol == 0 )
	/* max volume is 0x38 OR MAYBE 0x3f */
	
	//if(left_vol == 0 )
	//	left_vol = 0x30;
	//else
		left_vol  = 0x4f * left_vol  / 100;
	
	//if(right_vol == 0)
	//	right_vol = 0x30;	
	//else
		right_vol = 0x4f * right_vol / 100;
	
	left_vol += 0x0b0;
	right_vol += 0x0b0;
	printk("left_vol = %04x ,right_vol = %04x \n",left_vol,right_vol);
       // voldata = (left_vol  << 8) | right_vol;
	//printk("%x\n",voldata);i2c_smbus_write_byte_data
	if (i2c_smbus_write_word_data(tlv320_client, 2 << 1, left_vol) < 0) {
                printk("daca: failed to set volume \n");
                return -1;
        }
	if (i2c_smbus_write_word_data(tlv320_client, 3 << 1, right_vol) < 0) {
		printk("daca: failed to set volume \n");
		return -1;
	}
	//i2c_smbus_write_byte_data(tlv320_client, 2 << 1, left_vol);
	//i2c_smbus_write_byte_data(tlv320_client, 3 << 1, right_vol);
        return 0;
}

void tlv320_get_volumes(uint *vol) {
        *vol =  (cur_left_vol << 8) | cur_right_vol;
	
}

static int snd_tlv320_mixer_oss_open(struct inode *inode, struct file *file)
{
	if(tlv320_client == NULL)
		tlv320_client = tlv_get_i2c_clinet();	

	return 0;
}

static int snd_tlv320_mixer_oss_release(struct inode *inode, struct file *file)
{

	return 0;
}

#define SET_VOLUME 1
#define GET_VOLUME 2

static int snd_tlv320_mixer_oss_ioctl(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;

	int tmp;
	switch (cmd) {
		case SET_VOLUME://volume :
			tmp = *((int __user *) argp);
			tlv320_set_volumes(tmp,tmp);
		return 0;
		case GET_VOLUME:
			tlv320_get_volumes(&tmp);
			*((int __user *) arg) = tmp; 
		return 0;
	}
	
	return -ENXIO;
}

static struct file_operations tlv320_fops =
{
	.owner =	"sound",
	.open =		snd_tlv320_mixer_oss_open,
	.release =	snd_tlv320_mixer_oss_release,
	.ioctl =	snd_tlv320_mixer_oss_ioctl,
};


static int  tlv320_init(void)
{

	if (register_chrdev(234, "tmixer", &tlv320_fops)) {
		printk(KERN_ERR "tlv320: Unable to get major %d for tlv320\n"
		, 14);
		return -EIO;
	}
	printk("Tlv320 Mixer driver version 1.0\n");

	return 0;		
	
}

static void __exit  tlv320_exit(void)
{
	if (unregister_chrdev(TLV320_MAJOR, "tmixer")) {
		printk(KERN_ERR "lv320: Unable to release major %d for tlv320\n", TLV320_MAJOR);
		return;
	}
}



module_init( tlv320_init);
module_exit( tlv320_exit);

MODULE_AUTHOR("Ace Yang");
MODULE_LICENSE("GPL");
