/*
 *	Description: 
 *
 *
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include "matrix500.h"

#define PROGRAM "update"
#define VERSION "0.1.1"

static const char *exe_name;
static const char *file_name;
static int quiet; /* true -- don't output progress */
static struct lcd_display display;
void display_help()
{
    printf( "Usage: %s [OPTION] filename\n"
            "\n"
            "  -q, --quiet    		don't display progress messages\n"
	    "      --silent   		same as --quiet\n"
            "      --help     		display this help and exit\n"
            "      --version  		output version information and exit\n"
            "      --lighton/lightoff 	Turn ON/OFF backlight\n"
            "      --ledon/ledoff 		Turn ON/OFF LCD \n"
            "      --clear			Clear display \n"
            "      --test			Display demo pattern\n"
            , exe_name);
    exit(0);
}
    

void display_version()
{
    printf( PROGRAM " " VERSION "\n"
            "\n"
            "Copyright (C) 2006 Artila Electronics Ltd\n"
	    "\n"
            PROGRAM " comes with NO WARRANTY\n"
            "to the extent permitted by law.\n"
            "\n"
            "You may redistribute copies of " PROGRAM "\n"
            "under the terms of the GNU General Public Licence.\n"
            "See the file `COPYING' for more information.\n"                  
	);
    exit(0);
}

void process_options( int argc, char *argv[] )
{
    int error=0, fd=open("/dev/lcd",O_RDWR);;

    exe_name=argv[0];

    for(;;) {
        int option_index = 0;
        static const char* short_options="q";
        static const struct option long_options[] = {
            {"help", no_argument, 0, 0},
            {"version", no_argument, 0, 0},
            {"lighton", no_argument, 0, 0},
            {"lightoff", no_argument, 0, 0},
            {"lcdon", no_argument, 0, 0},
            {"lcdoff", no_argument, 0, 0},
            {"clear", no_argument, 0, 0},
            {"test", no_argument, 0, 0},
	    {"quiet", no_argument, 0, 'q'},
	    {"silent", no_argument, 0, 'q'},

            {0, 0, 0, 0},
        };

        int c=getopt_long( argc, argv, short_options,
                           long_options, &option_index );
        if( c==EOF ) {
            break;
        }

        switch( c ) {
        case 0 :
            switch( option_index ) {
            case 0 :
            	display_help();
                break;
            case 1 :
                display_version();
                break;
            case 2:
            	ioctl(fd, LCD_Back_On);
            	break;    
            case 3:
            	ioctl(fd, LCD_Back_Off);
            	break;
            case 4:
            	ioctl(fd, LCD_On);
            	break;
	    case 5:
	    	ioctl(fd, LCD_Off);
	    	break;            			
            case 6:
            	ioctl(fd, LCD_Clear);
            	break;
            case 7:
            	display.size1 = LCD_CHARS_PER_LINE;
		display.size2 = LCD_CHARS_PER_LINE;
		strcpy(display.line1, "Artila Electronics");
		strcpy(display.line2, "  Matrix520 Demo. ");
		ioctl(fd, LCD_Write, &display);
            	break;			
            }
            break;
	case 'q' :
	    quiet=1;
	    break;
        case '?' :
            error=1;
            break;
        }
    }
    close(fd);
    
    /*if( optind==argc ) {
	fprintf( stderr, "%s: no options specified\n", exe_name );
        error=1;
    }*/
    if( error || argc ==1 ) {
        fprintf( stderr, "Try `%s --help' for more information.\n", exe_name );
        exit( 1 );
    }

    file_name = argv[optind];
  
}


int main( int argc,char *argv[] )
{
	int i, fd;
	
	process_options( argc, argv );
	
	if(file_name ){	
		fd = open(file_name, O_RDONLY);
		i = read(fd, display.line1,LCD_CHARS_PER_LINE); 	
		display.size1 = i;
		read(fd, display.line2, 1);
		i = read(fd, display.line2,LCD_CHARS_PER_LINE); 
		display.size2 = i;
		close(fd);
		fd = open("/dev/lcd",O_RDWR);
		ioctl(fd, LCD_Write, &display);
		close(fd);
	}
	return 0;
}
