/*
 *  linux/drivers/char/at91_serial.c
 *
 *  Driver for Atmel AT91RM9200 Serial ports
 *
 *  Copyright (C) 2003 Rick Bronson
 *
 *  Based on drivers/char/serial_sa1100.c, by Deep Blue Solutions Ltd.
 *  Based on drivers/char/serial.c, by Linus Torvalds, Theodore Ts'o.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/tty.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/kd.h>
#include <linux/sysrq.h>
#include <linux/tty_flip.h>
//alan
#include <linux/pci.h>

#include <asm/arch/AT91RM9200_USART.h>
#include <asm/mach/serial_at91rm9200.h>
#include <asm/arch/board.h>
#include <asm/arch/pio.h>

#undef TLIA_BOARD

#if defined(CONFIG_SERIAL_AT91_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
#define SUPPORT_SYSRQ
#endif

#include <linux/serial_core.h>

#ifdef CONFIG_SERIAL_AT91_TTYAT

/* Use device name ttyAT, major 240 and minor 0-4.  This is necessary if we
 * should coexist with the 8250 driver, such as if we have an external 16C550
 * UART. */
#define SERIAL_AT91_MAJOR	240
#define MINOR_START		0
#define AT91_DEVICENAME		"ttyAT"

#else

/* Use device name ttyS, major 4, minor 64-68.  This is the usual serial port
 * name, but it is legally reserved for the 8250 driver. */
#define SERIAL_AT91_MAJOR	TTY_MAJOR
#define MINOR_START		64
#define AT91_DEVICENAME		"ttyS"

#endif

#define AT91C_VA_BASE_DBGU	((unsigned long) &(AT91_SYS->DBGU_CR))
#define AT91_ISR_PASS_LIMIT	256

#define UART_PUT_CR(port,v)	((AT91PS_USART)(port)->membase)->US_CR = v
#define UART_GET_MR(port)	((AT91PS_USART)(port)->membase)->US_MR
#define UART_PUT_MR(port,v)	((AT91PS_USART)(port)->membase)->US_MR = v
#define UART_PUT_IER(port,v)	((AT91PS_USART)(port)->membase)->US_IER = v
#define UART_PUT_IDR(port,v)	((AT91PS_USART)(port)->membase)->US_IDR = v
#define UART_GET_IMR(port)	((AT91PS_USART)(port)->membase)->US_IMR
#define UART_GET_CSR(port)	((AT91PS_USART)(port)->membase)->US_CSR
#define UART_GET_CHAR(port)	((AT91PS_USART)(port)->membase)->US_RHR
#define UART_PUT_CHAR(port,v)	((AT91PS_USART)(port)->membase)->US_THR = v
#define UART_GET_BRGR(port)	((AT91PS_USART)(port)->membase)->US_BRGR
#define UART_PUT_BRGR(port,v)	((AT91PS_USART)(port)->membase)->US_BRGR = v
#define UART_PUT_RTOR(port,v)	((AT91PS_USART)(port)->membase)->US_RTOR = v

#define UART_GET_CR(port)	((AT91PS_USART)(port)->membase)->US_CR		// is write-only

 /* PDC registers */
#define UART_PUT_PTCR(port,v)	((AT91PS_USART)(port)->membase)->US_PTCR = v
#define UART_PUT_RPR(port,v)	((AT91PS_USART)(port)->membase)->US_RPR = v
#define UART_PUT_RCR(port,v)	((AT91PS_USART)(port)->membase)->US_RCR = v
#define UART_GET_RCR(port)	((AT91PS_USART)(port)->membase)->US_RCR
#define UART_PUT_RNPR(port,v)	((AT91PS_USART)(port)->membase)->US_RNPR = v
#define UART_PUT_RNCR(port,v)	((AT91PS_USART)(port)->membase)->US_RNCR = v
//alan
#define UART_PUT_TPR(port,v)	((AT91PS_USART)(port)->membase)->US_TPR = v
#define UART_GET_TPR(port)	((AT91PS_USART)(port)->membase)->US_TPR
#define UART_PUT_TCR(port,v)	((AT91PS_USART)(port)->membase)->US_TCR = v
#define UART_GET_TCR(port)	((AT91PS_USART)(port)->membase)->US_TCR

#define UART_PUT_TNPR(port,v)	((AT91PS_USART)(port)->membase)->US_TNPR = v
#define UART_GET_TNPR(port)	((AT91PS_USART)(port)->membase)->US_TNPR
#define UART_PUT_TNCR(port,v)	((AT91PS_USART)(port)->membase)->US_TNCR = v
#define UART_GET_TNCR(port)	((AT91PS_USART)(port)->membase)->US_TNCR
#define NORMAL_UART_DMA_SIZE	50
#define UART_RECEIVER_TIMEOUT	30	/* in bit-times */
#define MATRIX500_GET_UART_TYPE	0xe001  /* alan */
#define MATRIX500_SET_UART_TYPE	0xe002	
#define MATRIX500_GET_UART_MODE9	0xe003  /* alan */
#define MATRIX500_SET_UART_MODE9	0xe004
#define MATRIX500_UART_SEND_BREAK	0xe005  /* ace */
#define MATRIX500_UART_SET_BAUD		0xe006 
#define MATRIX500_UART_SET_OVER		0xe007 
#define MATRIX500_UART_SET_SYNC		0xe008 
#define MATRIX500_UART_GET_OVER		0xe009 
#define MATRIX500_UART_GET_SYNC		0xe00a
#define MATRIX500_UART_TYPE_232		232
#define MATRIX500_UART_TYPE_422		422
#define MATRIX500_UART_TYPE_485		485
#define MATRIX500_UART_TYPE_TTL		0xe001

static int (*at91_open)(struct uart_port *);
static void (*at91_close)(struct uart_port *);


#ifdef SUPPORT_SYSRQ
static struct console at91_console;
#endif


//alan
static struct at91_serial_port at91_ports[AT91C_NR_UART];
//alan
struct at91_serial_port {
	struct uart_port	uart;
	
        dma_addr_t 	pdc_tx, pdc_rx;           	/* HW address for current transfer */
        unsigned char 	*driver_tx, *driver_rx;   	/* kernel address for current transfer */
	u_int16_t	dma_buffer_size;
        u_int16_t	rts;
	u_int16_t	dtr;
	u_int16_t	uart_type;
	u_int16_t	uart_mode9;
	u_int32_t	cached_CSR;
	unsigned int	uart_baud;//Ace
	unsigned int	quot;//Ace
	int 		sync;
	int 		over;
};

/*
 * at91_find_port()
 *
 * Return matching AT91 UART data structure, return NULL
 * if not found.
 */

struct at91_serial_port *at91_find_port(struct uart_port *port)
{
	int i;
	struct at91_serial_port *at91_port;

	at91_port = NULL;
	for (i = 0; i < AT91C_NR_UART; i++) {
		if ( &(at91_ports[i].uart) == port) {
			at91_port = &(at91_ports[i]);
		}
	}

	return at91_port;
}
/*
 * Return TIOCSER_TEMT when transmitter FIFO and Shift register is empty.
 */
static u_int at91_tx_empty(struct uart_port *port)
{
	return (UART_GET_CSR(port) & AT91C_US_TXEMPTY) ? TIOCSER_TEMT : 0;

}

/*
 * Set state of the modem control output lines
 */
static void at91_set_mctrl(struct uart_port *port, u_int mctrl)
{
	unsigned int control = 0;
	struct at91_serial_port *at91_port;
	at91_port = at91_find_port(port);
	if(at91_port->uart_type == MATRIX500_UART_TYPE_232){		


	
	if (mctrl & TIOCM_RTS){
		control |= AT91C_US_RTSEN;
		at91_port->rts = 1;//ace 1 to 0
	}
	else{
		control |= AT91C_US_RTSDIS;
		at91_port->rts = 0;//ace 0 to 1
	}

	if (mctrl & TIOCM_DTR){
		control |= AT91C_US_DTREN;
		at91_port->dtr = 1;//ace 1 to 0
	}
	else{
		control |=  AT91C_US_DTRDIS;
		at91_port->dtr = 0;//ace 0 to 1
	}

	}


	UART_PUT_CR(port,control);
}

/*
 * Get state of the modem control input lines
 */
static u_int at91_get_mctrl(struct uart_port *port)
{
	
	unsigned int ret = 0 , status;
	struct at91_serial_port *at91_port;
	status = UART_GET_CSR(port);
	at91_port = at91_find_port(port);
	
	//Ace ttyS2 is modem mode .
	if(port->mapbase ==  AT91C_VA_BASE_US1) {
		if (!(status & AT91C_US_DCD))
			ret |= TIOCM_CD;
		if (!(status & AT91C_US_CTS))  
			ret |= TIOCM_CTS;
		if (!(status & AT91C_US_DSR))
			ret |= TIOCM_DSR;
		if (!(status & AT91C_US_RI))
			ret |= TIOCM_RI;
		if(at91_port->rts)
			ret |= TIOCM_RTS;
		if(at91_port->dtr)
			ret |= TIOCM_DTR;
	}else {
		if (status & AT91C_US_DCD)
			ret |= TIOCM_CD;
		if (!(status & AT91C_US_CTS)) //alan  
			ret |= TIOCM_CTS;
		if (status & AT91C_US_DSR)
			ret |= TIOCM_DSR;
		if (status & AT91C_US_RI)
			ret |= TIOCM_RI;
		if(at91_port->rts)
			ret |= TIOCM_RTS;
		if(at91_port->dtr)
			ret |= TIOCM_DTR;
	}


	return ret;
}

/*
 * Stop transmitting.
 */
static void at91_stop_tx(struct uart_port *port)
{
	UART_PUT_IDR(port, AT91C_US_TXRDY);
	port->read_status_mask &= ~AT91C_US_TXRDY;
}

/*
 * Start transmitting.
 */
static void at91_start_tx(struct uart_port *port)
{
	unsigned long flags;

	local_irq_save(flags);
	port->read_status_mask |= AT91C_US_TXRDY;
	UART_PUT_IER(port, AT91C_US_TXRDY);
	local_irq_restore(flags);

}

/*
 * Stop receiving - port is in process of being closed.
 */
static void at91_stop_rx(struct uart_port *port)
{
	UART_PUT_IDR(port, AT91C_US_ENDRX | AT91C_US_RXBUFF | AT91C_US_TIMEOUT);
}

/*
 * Enable modem status interrupts
 */
static void at91_enable_ms(struct uart_port *port)
{
	UART_PUT_IER(port, AT91C_US_RIIC | AT91C_US_DSRIC | AT91C_US_DCDIC | AT91C_US_CTSIC);
}

/*
 * Control the transmission of a break signal
 */
static void at91_break_ctl(struct uart_port *port, int break_state)
{
	if (break_state != 0)
		UART_PUT_CR(port, AT91C_US_STTBRK);	/* start break */
	else
		UART_PUT_CR(port, AT91C_US_STPBRK);	/* stop break */
}




/*
 * Characters received (called from interrupt handler)
 *
 * The problem with DMA is that we have no idea
 * which character(s) caused a parity error, or when
 * exactly a framing error or overflow error occured.
 *
 * Basic flow:
 * 	Check for errors
 * 	copy out the characters from the DMA buffers
 * 	reset the DMA registers (to indicate that they are empty)
 * 	restart receiving
 */
static void at91_pdc_rx_chars(struct at91_serial_port *port, struct pt_regs *regs)
{
	int i;
	struct uart_port *uart = &port->uart;
	struct tty_struct *tty = uart->info->tty;
	unsigned int ch, flg;
	u_int16_t rx_count;

	flg = TTY_NORMAL;
	
	port->cached_CSR = UART_GET_CSR(uart);//ace
	
	if (port->cached_CSR & (AT91C_US_PARE | AT91C_US_FRAME | AT91C_US_OVRE)) {
		UART_PUT_CR(uart, AT91C_US_RSTSTA);	/* clear error */
		if (port->cached_CSR & (AT91C_US_OVRE)) {
			uart->icount.overrun++;
			flg = TTY_OVERRUN;
		}
		if (port->cached_CSR & (AT91C_US_PARE)) {
			uart->icount.parity++;
			flg = TTY_PARITY;
		}
		else if (port->cached_CSR & (AT91C_US_FRAME)) {
			uart->icount.frame++;
			flg = TTY_FRAME;
		}

#ifdef SUPPORT_SYSRQ
	uart->sysrq = 0;
#endif
	}

	UART_PUT_PTCR(uart, AT91C_PDC_RXTDIS);  // disable receive

	/* At this point, RX Next pointer and counter have
	 * been shifted into RX pointer and counter. */

	rx_count = port->dma_buffer_size - (UART_GET_RCR(uart));
	//Artila Ace For 9 bit .
	if(port->uart_mode9)
		rx_count = rx_count*2;

	if (rx_count != 0) {
		uart->icount.rx += rx_count;
		if ((tty->flip.count) + rx_count < TTY_FLIPBUF_SIZE) {
			for (i = 0; i < rx_count; i++) {
				ch = port->driver_rx[i];
				if (!uart_handle_sysrq_char(uart, ch, regs)) {
					*tty->flip.char_buf_ptr++ = ch;
					*tty->flip.flag_buf_ptr++ = flg;
				}
			}
			tty->flip.count += rx_count;
		}

	}
	
	// reset buffer to empty
	UART_PUT_RPR(uart, port->pdc_rx);
	UART_PUT_RCR(uart, port->dma_buffer_size);

	// enable receive
	UART_PUT_PTCR(uart, AT91C_PDC_RXTEN);  

	tty_flip_buffer_push(tty);
}


/*
 * Transmit characters (called from interrupt handler)
 */
static void at91_tx_chars(struct at91_serial_port *port)
{
	struct uart_port *uart = &port->uart;
	struct circ_buf *xmit = &uart->info->xmit;
	short sbuf;
	if (uart->x_char) {

		UART_PUT_CHAR(uart, uart->x_char);
		uart->icount.tx++;
		uart->x_char = 0;
		return;
	}

	if (uart_circ_empty(xmit) || uart_tx_stopped(uart)) {
		at91_stop_tx(uart);
		return;
	}

	if (UART_GET_CSR(uart) & AT91C_US_TXRDY) {
		if(port->uart_mode9) {
			sbuf = (xmit->buf[xmit->tail+1] << 8) + xmit->buf[xmit->tail];
			UART_PUT_CHAR(uart, sbuf);
			xmit->tail = (xmit->tail + 2) & (UART_XMIT_SIZE - 1);
			uart->icount.tx++;
		}else {
			UART_PUT_CHAR(uart, xmit->buf[xmit->tail]);
			xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
			uart->icount.tx++;
		}

	}

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(uart);

	if (uart_circ_empty(xmit)) {
		at91_stop_tx(uart);
	}
}


/*
 * Interrupt handler
 *
 */
static irqreturn_t at91_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct at91_serial_port *port = dev_id;
	struct uart_port *uart = &port->uart;
	unsigned int pending, pass_counter = 0;
	
	port->cached_CSR = UART_GET_CSR(uart);
	pending = port->cached_CSR & uart->read_status_mask;
	while (pending) {
		if (pending & (AT91C_US_ENDRX | AT91C_US_RXBUFF | AT91C_US_TIMEOUT)) {
			at91_pdc_rx_chars(port, regs);
			UART_PUT_CR(uart, AT91C_US_STTTO);
		}

		/* Clear the relevent break bits */
		if (pending & AT91C_US_RXBRK) {
			UART_PUT_CR(uart, AT91C_US_RSTSTA);
			uart->icount.brk++;
#ifdef SUPPORT_SYSRQ
			if (uart->line == at91_console.index && !uart->sysrq) {
				uart->sysrq = jiffies + HZ*5;
			}
#endif
		}

		if (pending & AT91C_US_RIIC) uart->icount.rng++;
		if (pending & AT91C_US_DSRIC) uart->icount.dsr++;
		/* It appears that when the AT91C_US_DCD is '1',
                 * this means that the DCD pin is being driven low (active).  */
		if (pending & AT91C_US_DCDIC) {
			uart_handle_dcd_change(uart, !(port->cached_CSR & AT91C_US_DCD));//ace
		}

                /* It appears that when the AT91C_US_CTS is '1',
                 * this means that the CTS pin is being driven low (active).  */
		if (pending & AT91C_US_CTSIC) {
			uart_handle_cts_change(uart, (port->cached_CSR & AT91C_US_CTS) ? 0 : 1 );
		}

		/* Not sure if this is still needed.  I believe that by accessing
		 * the CSR in only one location (this ISR), we have eliminated the
		 * bug where we were missing CTS transitions. */

		if (pending & (AT91C_US_RIIC | AT91C_US_DSRIC | AT91C_US_DCDIC | AT91C_US_CTSIC))
			wake_up_interruptible(&uart->info->delta_msr_wait);


		if (pending & AT91C_US_TXRDY) {
			at91_tx_chars(port);
		}

		if (pass_counter++ > AT91_ISR_PASS_LIMIT)
			break;

		port->cached_CSR = UART_GET_CSR(uart);
		pending = port->cached_CSR & uart->read_status_mask;
	}
	return IRQ_HANDLED;
}


/*
 * Perform initialization and enable port for reception
 */
static int at91_startup(struct uart_port *port)
{
	int retval;
	//alan
	struct at91_serial_port *at91_port;
	
	at91_port = at91_find_port(port);

	if (at91_port == NULL) {
		printk("at91_serial: at91_startup, membase: %x  ", ((unsigned int)(port->membase)));
		printk("- can't find matching port struct.\n");
		return 1;
	}

	at91_port->driver_rx 	  = pci_alloc_consistent(NULL, at91_port->dma_buffer_size, &(at91_port->pdc_rx));
	
	if (at91_port->driver_rx == NULL) {
		printk("at91_serial: at91_startup - memory allocation failed for PDC buffers.\n");
		return 1;
	}
	else {
		UART_PUT_RPR(port, at91_port->pdc_rx);
		UART_PUT_RCR(port, at91_port->dma_buffer_size);
	}

	/* 
	 * Ensure that no interrupts are enabled otherwise when 
	 * request_irq() is called we could get stuck trying to 
	 * handle an unexpected interrupt
	 */
	UART_PUT_IDR(port, -1);

	/*
	 * Allocate the IRQ
	 */
	retval = request_irq(port->irq, at91_interrupt, SA_SHIRQ, "at91_serial", port);
	if (retval) {
		printk("at91_serial: at91_startup - Can't get irq\n");
		return retval;
	}

	/*
	 * If there is a specific "open" function (to register
	 * control line interrupts)
	 */
	if (at91_open) {
		retval = at91_open(port);
		if (retval) {
			free_irq(port->irq, port);
			return retval;
		}
	}
	//alan
	/* Enable peripheral clock if required */
	if (port->irq != AT91C_ID_SYS)
		AT91_SYS->PMC_PCER = 1 << port->irq;



	port->read_status_mask = AT91C_US_TXRDY | AT91C_US_OVRE
			| AT91C_US_FRAME | AT91C_US_PARE | AT91C_US_RXBRK
			| AT91C_US_ENDRX | AT91C_US_RXBUFF | AT91C_US_TIMEOUT
			| AT91C_US_CTSIC ;
	/*
	 * Finally, clear and enable interrupts
	 */
	UART_PUT_RTOR(port, UART_RECEIVER_TIMEOUT);
	UART_PUT_IDR(port, -1);
	UART_PUT_PTCR(port, AT91C_PDC_RXTEN);			/* enable PDC receive */
	UART_PUT_CR(port, AT91C_US_TXEN | AT91C_US_RXEN | AT91C_US_STTTO);	/* enable xmit, rcvr, and rcvr timeout  */
	UART_PUT_IER(port, AT91C_US_ENDRX | AT91C_US_RXBUFF | AT91C_US_TIMEOUT);
	return 0;
}

/*
 * Disable the port
 */
static void at91_shutdown(struct uart_port *port)
{
	struct at91_serial_port *at91_port;
	unsigned long flags ;
	
	local_irq_save(flags);
	// completely disable PDC

	UART_PUT_PTCR(port, AT91C_PDC_RXTDIS | AT91C_PDC_TXTDIS);
	at91_stop_rx(port);//Ace
	at91_stop_tx(port);//Ace
	UART_PUT_RPR(port, 0);
	UART_PUT_RCR(port, 0);
	UART_PUT_RNPR(port, 0);
	UART_PUT_RNCR(port, 0);
	UART_PUT_TPR(port, 0);
	UART_PUT_TCR(port, 0);
	UART_PUT_TNPR(port, 0);
	UART_PUT_TNCR(port, 0);

	at91_port = at91_find_port(port);
	pci_free_consistent(NULL, at91_port->dma_buffer_size, at91_port->driver_rx, at91_port->pdc_rx);

	/*
	 * Disable all interrupts, port and break condition.
	 */
	UART_PUT_CR(port, AT91C_US_RSTSTA);//ace
	UART_PUT_IDR(port, -1);
	/*
	 * Free the interrupt
	 */
	free_irq(port->irq, port);

	/*
	 * If there is a specific "close" function (to unregister
	 * control line interrupts)
	 */
	if (at91_close)
		at91_close(port);

	/* Disable peripheral clock if required */
	if (port->irq != AT91C_ID_SYS)
		AT91_SYS->PMC_PCDR = 1 << port->irq;
	local_irq_restore(flags);
}


/*
 * Power / Clock management.
 */
static void at91_serial_pm(struct uart_port *port, unsigned int state, unsigned int oldstate)
{
	switch (state) {
		case 0:
			/*
			 * Enable the peripheral clock for this serial port.
			 * This is called on uart_open() or a resume event.
			 */
			AT91_SYS->PMC_PCER = 1 << port->irq;
			break;
		case 3:
			/*
			 * Disable the peripheral clock for this serial port.
			 * This is called on uart_close() or a suspend event.
			 */
			if (port->irq != AT91C_ID_SYS)			/* is this a shared clock? */
				AT91_SYS->PMC_PCDR = 1 << port->irq;
			break;
		default:
			printk(KERN_ERR "at91_serial: unknown pm %d\n", state);
	}
}



/* alan */
static int at91_ioctl(struct uart_port *port, unsigned int cmd, unsigned long arg)
{
	int __user *type = (int __user *)arg;
	struct at91_serial_port *at91_port;
	unsigned long flags ;
	//printk("at91_ioctl:0x%x\n", cmd);
	at91_port = at91_find_port(port);
	switch(cmd){
		case MATRIX500_GET_UART_TYPE:
		
			*type = at91_port->uart_type; 
				
		break;
		case MATRIX500_SET_UART_TYPE:
			if(port->mapbase ==  AT91C_VA_BASE_US0 ) {
				if(*type==MATRIX500_UART_TYPE_232) {
					at91_port->uart_type = *type;
					at91_set_gpio_output(AT91_PIN_PD13,1);
					at91_set_gpio_output(AT91_PIN_PD14,1);
				}
				else if(*type==MATRIX500_UART_TYPE_422 )  
				{
					at91_port->uart_type = *type;
					at91_set_gpio_output(AT91_PIN_PD13,0);	
					at91_set_gpio_output(AT91_PIN_PD14,1);
				}else if(*type==MATRIX500_UART_TYPE_485 )  {
					at91_port->uart_type = *type;
					at91_set_gpio_output(AT91_PIN_PD13,0);	
					at91_set_gpio_output(AT91_PIN_PD14,0);
				}
			}	
		break;
		case MATRIX500_GET_UART_MODE9:
			*type = at91_port->uart_mode9; 
			
		break;
		case MATRIX500_SET_UART_MODE9:
			
			at91_port->uart_mode9 = *type;
			
		break;
		// Ace
		case  MATRIX500_UART_SEND_BREAK:
			local_irq_save(flags);
			UART_PUT_CR( port , AT91C_US_STTBRK );
			if(*type < 1000)
				udelay(*type);
			else
				mdelay(*type/1000);
			UART_PUT_CR( port ,  AT91C_US_STPBRK );
			local_irq_restore(flags);
		break;
		case  MATRIX500_UART_SET_BAUD:
			
			at91_port->uart_baud = *type;
			
		break;
		case  MATRIX500_UART_SET_OVER:
		{
			if(*type)
				at91_port->over = 1;
			else
				at91_port->over = 0;
		}
		break;
		case  MATRIX500_UART_SET_SYNC:
		{
			if(*type)
				at91_port->sync = 1;
			else
				at91_port->sync = 0;
		}
		break;
		case MATRIX500_UART_GET_OVER:
			*type = at91_port->over; 
		break;
		case MATRIX500_UART_GET_SYNC:
			*type = at91_port->sync; 
		break;
		case 0xf001:
			local_irq_save(flags);
			UART_PUT_IER(port, AT91C_US_RXBRK);
			local_irq_restore(flags);
		break;
		case KDMKTONE:
			{
			unsigned int ticks, count, i,j;
			//arg = (duration<<16)|(freq?(1193180/freq):0) 
			/*
		 	* Generate the tone for the appropriate number of ticks.
		 * If the time is zero, turn off sound ourselves.
		 */
			ticks = HZ * ((arg >> 16) & 0xffff) / 1000;
			count = ticks ? (arg & 0xffff) : 0;
			for(i=0; i<= ticks;i++){
				AT91_SYS->PIOB_SODR = AT91C_PIO_PB6;
				for(j=0;j<count;j++);;
				AT91_SYS->PIOB_CODR = AT91C_PIO_PB6;
				for(j=0;j<count*2;j++);;		
			}
			}
		break;
		default:
		return -ENOIOCTLCMD;
		
	}
	return 0;//-ENOIOCTLCMD;
}


static struct uart_ops at91_pops;		/* forward declaration */


/*
 * Change the port parameters
 */
static void at91_set_termios(struct uart_port *port, struct termios * termios, struct termios * old)
{
	unsigned long flags;
	unsigned int mode, imr ,quot, baud;
	struct at91_serial_port *at91_port;
	
	at91_port = at91_find_port(port);
	
	if(at91_port->uart_baud != 0 && old == NULL) { //Ace
		baud = at91_port->uart_baud;
		termios->c_cflag |= B4000000;
	}
	else {
		at91_port->uart_baud = 0; 
		baud = uart_get_baud_rate(port, termios, old, 0, port->uartclk/16); 
	}
	quot = uart_get_divisor(port, baud);
	if(port->mapbase != AT91C_VA_BASE_DBGU) {
		if(at91_port->sync == 1) {
			quot = quot * 16;
			mode |=  AT91C_US_SYNC;
		}else {
			if(at91_port->over == 1) {
				quot = quot * 2;
				mode |= AT91C_US_OVER;
			}
		}
		if (quot > 65535) {	/* BRGR is 16-bit, so switch to slower clock */
			quot /= 8;
			mode |= AT91C_US_CLKS_FDIV1;
		}
		
	}	
	uart_update_timeout(port, termios->c_cflag, baud);
	
	/* Get current mode register */
	mode = UART_GET_MR(port) & ~(AT91C_US_CHRL | AT91C_US_NBSTOP | AT91C_US_PAR );
				
	if((UART_GET_MR(port) & AT91C_US_SYNC) || (UART_GET_MR(port) & AT91C_US_SYNC)) //Ace
		mode = mode & ~(AT91C_US_SYNC | AT91C_US_OVER);

	/* byte size */
	switch (termios->c_cflag & CSIZE) {
	case CS5:
		mode |= AT91C_US_CHRL_5_BITS;
		break;
	case CS6:
		mode |= AT91C_US_CHRL_6_BITS;
		break;
	case CS7:
		mode |= AT91C_US_CHRL_7_BITS;
		break;
	default:
		if(at91_port->uart_mode9 == 0)
			mode |= AT91C_US_CHRL_8_BITS;
		break;
	}
	
	if(at91_port->uart_mode9 > 0)
		mode |= AT91C_US_MODE9;
	else	
		mode &= ~AT91C_US_MODE9;

	/* stop bits */
	if (termios->c_cflag & CSTOPB)
		mode |= AT91C_US_NBSTOP_2_BIT;

	/* parity */
	if (termios->c_cflag & PARENB) {
		if (termios->c_cflag & CMSPAR) {			/* Mark or Space parity */
			if (termios->c_cflag & PARODD)
				mode |= AT91C_US_PAR_MARK;
			else
				mode |= AT91C_US_PAR_SPACE;
		}
		else if (termios->c_cflag & PARODD)
			mode |= AT91C_US_PAR_ODD;
		else
			mode |= AT91C_US_PAR_EVEN;
	}
	else
		mode |= AT91C_US_PAR_NONE;

	port->read_status_mask |= AT91C_US_OVRE;
	if (termios->c_iflag & INPCK)
		port->read_status_mask |= AT91C_US_FRAME | AT91C_US_PARE;
	if (termios->c_iflag & (BRKINT | PARMRK))
		port->read_status_mask |= AT91C_US_RXBRK;

	/*
	 * Characters to ignore
	 */
	port->ignore_status_mask = 0;
	if (termios->c_iflag & IGNPAR)
		port->ignore_status_mask |= (AT91C_US_FRAME | AT91C_US_PARE);
	if (termios->c_iflag & IGNBRK) {
		port->ignore_status_mask |= AT91C_US_RXBRK;
		/*
		 * If we're ignoring parity and break indicators,
		 * ignore overruns too (for real raw support).
		 */
		if (termios->c_iflag & IGNPAR)
			port->ignore_status_mask |= AT91C_US_OVRE;
	}

	if (UART_ENABLE_MS(port, termios->c_cflag)){
			mode |= AT91C_US_USMODE_HWHSH;	
	}
	
	if(port->mapbase ==  AT91C_VA_BASE_US0){
		if(at91_port->uart_type != MATRIX500_UART_TYPE_232){
			if(at91_port->uart_type == MATRIX500_UART_TYPE_422){
				mode &= ~AT91C_US_USMODE_RS485;
				UART_PUT_CR(port,AT91C_US_RTSDIS);
			}
			else{
				mode |= AT91C_US_USMODE_RS485; 
			}	
		}
		else {
			
			mode &= ~AT91C_US_USMODE_RS485;
		}	
	}



	/* first, disable interrupts and drain transmitter */
	local_irq_save(flags);
	imr = UART_GET_IMR(port);	/* get interrupt mask */
	UART_PUT_IDR(port, -1);		/* disable all interrupts */
	local_irq_restore(flags);
	//while (!(UART_GET_CSR(port) & AT91C_US_TXEMPTY)) { barrier(); }//alan
	
	/* disable receiver and transmitter */
	UART_PUT_CR(port, AT91C_US_TXDIS | AT91C_US_RXDIS);
	
	/* set the parity, stop bits and data size */
	UART_PUT_MR(port, mode );//ALAN FOR TEST
	
	/* set the baud rate */
	UART_PUT_BRGR(port, quot);
	udelay(500);
	UART_PUT_RTOR(port, UART_RECEIVER_TIMEOUT);  // redundant?, alan
	UART_PUT_CR(port, AT91C_US_TXEN | AT91C_US_RXEN | AT91C_US_STTTO );//alan
	
	/* restore interrupts */
	UART_PUT_IER(port, imr);

	/* CTS flow-control and modem-status interrupts */
	if (UART_ENABLE_MS(port, termios->c_cflag))
		at91_pops.enable_ms(port);
	
}	

/*
 * Return string describing the specified port
 */
static const char *at91_type(struct uart_port *port)
{
	return (port->type == PORT_AT91RM9200) ? "AT91_SERIAL" : NULL;
}

/*
 * Release the memory region(s) being used by 'port'.
 */
static void at91_release_port(struct uart_port *port)
{
	release_mem_region(port->mapbase,
		(port->mapbase == AT91C_VA_BASE_DBGU) ? 512 : SZ_16K);
}

/*
 * Request the memory region(s) being used by 'port'.
 */
static int at91_request_port(struct uart_port *port)
{
	return request_mem_region(port->mapbase,
		(port->mapbase == AT91C_VA_BASE_DBGU) ? 512 : SZ_16K,
		"at91_serial") != NULL ? 0 : -EBUSY;

}

/*
 * Configure/autoconfigure the port.
 */
static void at91_config_port(struct uart_port *port, int flags)
{
	if (flags & UART_CONFIG_TYPE) {
		port->type = PORT_AT91RM9200;
		at91_request_port(port);
	}
}

/*
 * Verify the new serial_struct (for TIOCSSERIAL).
 */
static int at91_verify_port(struct uart_port *port, struct serial_struct *ser)
{
	int ret = 0;
	if (ser->type != PORT_UNKNOWN && ser->type != PORT_AT91RM9200)
		ret = -EINVAL;
	if (port->irq != ser->irq)
		ret = -EINVAL;
	if (ser->io_type != SERIAL_IO_MEM)
		ret = -EINVAL;
	if (port->uartclk / 16 != ser->baud_base)
		ret = -EINVAL;
	if ((void *)port->mapbase != ser->iomem_base)
		ret = -EINVAL;
	if (port->iobase != ser->port)
		ret = -EINVAL;
	if (ser->hub6 != 0)
		ret = -EINVAL;
	return ret;
}

static struct uart_ops at91_pops = {
	.tx_empty	= at91_tx_empty,
	.set_mctrl	= at91_set_mctrl,
	.get_mctrl	= at91_get_mctrl,
	.stop_tx	= at91_stop_tx,
	.start_tx	= at91_start_tx,
	.stop_rx	= at91_stop_rx,
	.enable_ms	= at91_enable_ms,
	.break_ctl	= at91_break_ctl,
	.startup	= at91_startup,
	.shutdown	= at91_shutdown,
	.set_termios	= at91_set_termios,
	.type		= at91_type,
	.release_port	= at91_release_port,
	.request_port	= at91_request_port,
	.config_port 	= at91_config_port,
	.verify_port 	= at91_verify_port,
	.pm		= at91_serial_pm,
	.ioctl		= at91_ioctl,		/* alan */
};

/*
 * We wrap our port structure around the generic uart_port.
 */



//static struct uart_port at91_ports[AT91C_NR_UART];

void __init at91_init_ports(void)
{
	static int first = 1;
	int i;

	if (!first)
		return;
	first = 0;

	for (i = 0; i < AT91C_NR_UART; i++) {
		at91_ports[i].uart.iotype	= SERIAL_IO_MEM;
		at91_ports[i].uart.flags     = ASYNC_BOOT_AUTOCONF;
		at91_ports[i].uart.uartclk   = at91_master_clock;
		at91_ports[i].uart.ops	= &at91_pops;
		at91_ports[i].uart.fifosize  = 1;
		at91_ports[i].uart.line	= i;
		at91_ports[i].uart_type = MATRIX500_UART_TYPE_232;
		at91_ports[i].uart_mode9 = 0;
		at91_ports[i].uart_baud = 0;//Ace
		at91_ports[i].sync = 0;//Ace
		at91_ports[i].over = 0;//Ace
		//alan
		if (i == 0) {
			/* The DBGU serial port (ttyS0) doesn't support
			 * RX timeout.  */
			at91_ports[i].dma_buffer_size = 1;
		}
		else {
			at91_ports[i].dma_buffer_size = NORMAL_UART_DMA_SIZE;
		}
 	}
}

void __init at91_register_uart_fns(struct at91rm9200_port_fns *fns)
{
	if (fns->enable_ms)
		at91_pops.enable_ms = fns->enable_ms;
	if (fns->get_mctrl)
		at91_pops.get_mctrl = fns->get_mctrl;
	if (fns->set_mctrl)
		at91_pops.set_mctrl = fns->set_mctrl;
	at91_open          = fns->open;
	at91_close         = fns->close;
	at91_pops.pm       = fns->pm;
	at91_pops.set_wake = fns->set_wake;
}

/*
 * Setup ports.
 */
void __init at91_register_uart(int idx, int port)
{
	
	if ((idx < 0) || (idx >= AT91C_NR_UART)) {
		printk(KERN_ERR "%s: bad index number %d\n", __FUNCTION__, idx);
		return;
	}

	switch (port) {
	case 0:
		at91_ports[idx].uart.membase = (void *) AT91C_VA_BASE_US0;
		at91_ports[idx].uart.mapbase = AT91C_VA_BASE_US0;
		at91_ports[idx].uart.irq     = AT91C_ID_US0;
		AT91_CfgPIO_USART0();
		break;
	case 1:
		at91_ports[idx].uart.membase = (void *) AT91C_VA_BASE_US1;
		at91_ports[idx].uart.mapbase = AT91C_VA_BASE_US1;
		at91_ports[idx].uart.irq     = AT91C_ID_US1;
		AT91_CfgPIO_USART1();
		break;
	case 2:
		at91_ports[idx].uart.membase = (void *) AT91C_VA_BASE_US2;
		at91_ports[idx].uart.mapbase = AT91C_VA_BASE_US2;
		at91_ports[idx].uart.irq     = AT91C_ID_US2;
		AT91_CfgPIO_USART2();

		break;
	case 3:
		at91_ports[idx].uart.membase = (void *) AT91C_VA_BASE_US3;
		at91_ports[idx].uart.mapbase = AT91C_VA_BASE_US3;
		at91_ports[idx].uart.irq     = AT91C_ID_US3;
		AT91_CfgPIO_USART3();
		break;
	case 4:
		at91_ports[idx].uart.membase = (void *) AT91C_VA_BASE_DBGU;
		at91_ports[idx].uart.mapbase = AT91C_VA_BASE_DBGU;
		at91_ports[idx].uart.irq     = AT91C_ID_SYS;
		AT91_CfgPIO_DBGU();
		break;
	default:
		printk(KERN_ERR  "%s : bad port number %d\n", __FUNCTION__, port);
	}
}

#ifdef CONFIG_SERIAL_AT91_CONSOLE

/*
 * Interrupts are disabled on entering
 */
static void at91_console_write(struct console *co, const char *s, u_int count)
{
	struct uart_port *port = &at91_ports[co->index].uart;//alan, at91_ports + co->index;
	unsigned int status, i, imr;

	/*
	 *	First, save IMR and then disable interrupts
	 */
	imr = UART_GET_IMR(port);	/* get interrupt mask */
	//UART_PUT_IDR(port, AT91C_US_RXRDY | AT91C_US_TXRDY);
	UART_PUT_IDR(port, AT91C_US_TXRDY | AT91C_US_ENDRX | AT91C_US_RXBUFF | AT91C_US_TIMEOUT);//alan
	/*
	 *	Now, do each character
	 */
	for (i = 0; i < count; i++) {
		do {
			status = UART_GET_CSR(port);
		} while (!(status & AT91C_US_TXRDY));
		UART_PUT_CHAR(port, s[i]);
		if (s[i] == '\n') {
			do {
				status = UART_GET_CSR(port);
			} while (!(status & AT91C_US_TXRDY));
			UART_PUT_CHAR(port, '\r');
		}
	}

	/*
	 *	Finally, wait for transmitter to become empty
	 *	and restore IMR
	 */
	do {
		status = UART_GET_CSR(port);
	} while (!(status & AT91C_US_TXRDY));
	UART_PUT_IER(port, imr);	/* set interrupts back the way they were */
}

/*
 * If the port was already initialised (eg, by a boot loader), try to determine
 * the current setup.
 */
static void __init at91_console_get_options(struct uart_port *port, int *baud, int *parity, int *bits)
{
	unsigned int mr, quot;

// TODO: CR is a write-only register
//	unsigned int cr;
//
//	cr = UART_GET_CR(port) & (AT91C_US_RXEN | AT91C_US_TXEN);
//	if (cr == (AT91C_US_RXEN | AT91C_US_TXEN)) {
//		/* ok, the port was enabled */
//	}

	mr = UART_GET_MR(port) & AT91C_US_CHRL;
	if (mr == AT91C_US_CHRL_8_BITS)
		*bits = 8;
	else
		*bits = 7;

	mr = UART_GET_MR(port) & AT91C_US_PAR;
	if (mr == AT91C_US_PAR_EVEN)
		*parity = 'e';
	else if (mr == AT91C_US_PAR_ODD)
		*parity = 'o';

	quot = UART_GET_BRGR(port);
	*baud = port->uartclk / (16 * (quot));
}

static int __init at91_console_setup(struct console *co, char *options)
{
	struct uart_port *port;
	int baud = 115200;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';

	/*
	 * Check whether an invalid uart number has been specified, and
	 * if so, search for the first available port that does have
	 * console support.
	 */
	/*
	 * Check whether an invalid uart number has been specified, and
	 * if so, search for the first available port that does have
	 * console support.
	 */
	if (co->index >= AT91C_NR_UART)
		co->index = 0;
	port = &at91_ports[co->index].uart;
	//port = uart_get_console(at91_ports, AT91C_NR_UART, co);

	/*
	 * Enable the serial console, in-case bootloader did not do it.
	 */
	AT91_SYS->PMC_PCER = 1 << port->irq;		/* enable clock */
	UART_PUT_IDR(port, -1);				/* disable interrupts */
	UART_PUT_CR(port, AT91C_US_RSTSTA | AT91C_US_RSTRX);
	UART_PUT_CR(port, AT91C_US_TXEN | AT91C_US_RXEN);
	
	if (options)
		uart_parse_options(options, &baud, &parity, &bits, &flow);
	else
		at91_console_get_options(port, &baud, &parity, &bits);

	return uart_set_options(port, co, baud, parity, bits, flow);
}

extern struct uart_driver at91_uart;

static struct console at91_console = {
	.name		= AT91_DEVICENAME,
	.write		= at91_console_write,
	.device		= uart_console_device,
	.setup		= at91_console_setup,
	.flags		= CON_PRINTBUFFER,
	.index		= -1,
	.data		= &at91_uart,
};

#define AT91_CONSOLE_DEVICE	&at91_console

static int  __init at91_console_init(void)
{
	at91_init_ports();

	at91_console.index = at91_console_port;
	register_console(&at91_console);
	return 0;
}
console_initcall(at91_console_init);

#else
#define AT91_CONSOLE_DEVICE	NULL
#endif

static struct uart_driver at91_uart = {
	.owner			= THIS_MODULE,
	.driver_name		= AT91_DEVICENAME,
	.dev_name		= AT91_DEVICENAME,
	.devfs_name		= AT91_DEVICENAME,
	.major			= SERIAL_AT91_MAJOR,
	.minor			= MINOR_START,
	.nr			= AT91C_NR_UART,
	.cons			= AT91_CONSOLE_DEVICE,
};

static int __init at91_serial_init(void)
{
	int ret, i;

	at91_init_ports();

	ret = uart_register_driver(&at91_uart);
	if (ret)
		return ret;

	if(console_drivers->name != NULL)//Ace
		uart_add_one_port(&at91_uart, &at91_ports[0].uart);
	else 
		AT91_CfgPIO_USART2_1();
	
	for (i = 1; i < AT91C_NR_UART; i++) {
		if (at91_serial_map[i] >= 0)
			uart_add_one_port(&at91_uart, &at91_ports[i].uart);
	}

	return 0;
}

static void __exit at91_serial_exit(void)
{
	int i;

	for (i = 0; i < AT91C_NR_UART; i++) {
 		if (at91_serial_map[i] >= 0)
			uart_remove_one_port(&at91_uart, &at91_ports[i].uart);
  	}

	uart_unregister_driver(&at91_uart);
}

module_init(at91_serial_init);
module_exit(at91_serial_exit);

MODULE_AUTHOR("Rick Bronson");
MODULE_DESCRIPTION("AT91 generic serial port driver");
MODULE_LICENSE("GPL");
