/*
   drivers/sound/at91rm9200ssc.c
   Linear codec driver for AT91RM9200 chip using SSC
   Copyright (C) 2004 Halina Nowak  (halina@7chips.com)
   AT91 subroutines are from ATMEL software

*/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/mm.h>

#include <asm/hardware.h>
#include <asm/setup.h>
#include <asm/mach-types.h>
#include <asm/irq.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>

#include <asm/arch/hardware.h>

#include <asm/arch/AT91RM9200_SSC.h>

#include "sound_config.h"
#include "dev_table.h"

#define OKI_7716_CODEC 1
#ifdef OKI_7716_CODEC

#define LOCAL_BUFFER 1

#define FSYNC 8000
/*
    FSYNC 8kHz, SSC_SCLK 832000
    14-bit data length with MSB sent first
    FSYNC is a negative pulse
    Status sync bits set on falling edge of FSYNC
*/
#define RCMR_PERIOD  51
#define CMR_DIV 18
#define SSC_SCLK  FSYNC*2*52

//#define SSC_BITS  14
#define SSC_BITS  16
#define RCMR_VAL  ((RCMR_PERIOD << 24) | AT91C_SSC_CKS_RK | AT91C_SSC_CKO_NONE | AT91C_SSC_START_RISE_RF)
#define RFMR_VAL  (AT91C_SSC_MSBF | AT91C_SSC_FSOS_NONE)

#define TCMR_VAL  ((RCMR_PERIOD <<24) | AT91C_SSC_CKS_DIV | AT91C_SSC_CKO_CONTINOUS | AT91C_SSC_CKI | AT91C_SSC_START_RISE_RF)
#define TFMR_VAL  ( AT91C_SSC_MSBF | AT91C_SSC_FSOS_POSITIVE )

#endif

#if 0
#define DBG(x...)   printk("AT91SSC :" x)
#define LDBG(x...)   printk(x)
#else
#define DBG(x...)   
#define LDBG(x...)   
#endif

/* *****************************************************************************
                SOFTWARE API FOR PDC
   ***************************************************************************** */
//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_SetNextRx
//* \brief Set the next receive transfer descriptor
//*----------------------------------------------------------------------------
 void AT91F_PDC_SetNextRx (
	AT91PS_SSC pSSC,     // \arg pointer to a PDC controller
	char *address,       // \arg address to the next bloc to be received
	unsigned int bytes)  // \arg number of bytes to be received
{
	pSSC->SSC_RNPR = (unsigned int) address;
	pSSC->SSC_RNCR = bytes;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_SetNextTx
//* \brief Set the next transmit transfer descriptor
//*----------------------------------------------------------------------------
 void AT91F_PDC_SetNextTx (
	AT91PS_SSC pSSC,       // \arg pointer to a PDC controller
	char *address,         // \arg address to the next bloc to be transmitted
	unsigned int bytes)    // \arg number of bytes to be transmitted
{
	pSSC->SSC_TNPR = (unsigned int) address;
	pSSC->SSC_TNCR = bytes;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_SetRx
//* \brief Set the receive transfer descriptor
//*----------------------------------------------------------------------------
 void AT91F_PDC_SetRx (
	AT91PS_SSC pSSC,       // \arg pointer to a PDC controller
	char *address,         // \arg address to the next bloc to be received
	unsigned int bytes)    // \arg number of bytes to be received
{
	pSSC->SSC_RPR = (unsigned int) address;
	pSSC->SSC_RCR = bytes;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_SetTx
//* \brief Set the transmit transfer descriptor
//*----------------------------------------------------------------------------
 void AT91F_PDC_SetTx (
	AT91PS_SSC pSSC,       // \arg pointer to a PDC controller
	char *address,         // \arg address to the next bloc to be transmitted
	unsigned int bytes)    // \arg number of bytes to be transmitted
{
	pSSC->SSC_TPR = (unsigned int) address;
	pSSC->SSC_TCR = bytes;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_EnableTx
//* \brief Enable transmit
//*----------------------------------------------------------------------------
 void AT91F_PDC_EnableTx (
	AT91PS_SSC pSSC )       // \arg pointer to a PDC controller
{
	pSSC->SSC_PTCR = AT91C_PDC_TXTEN;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_EnableRx
//* \brief Enable receive
//*----------------------------------------------------------------------------
 void AT91F_PDC_EnableRx (
	AT91PS_SSC pSSC )       // \arg pointer to a PDC controller
{
	pSSC->SSC_PTCR = AT91C_PDC_RXTEN;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_DisableTx
//* \brief Disable transmit
//*----------------------------------------------------------------------------
 void AT91F_PDC_DisableTx (
	AT91PS_SSC pSSC )       // \arg pointer to a PDC controller
{
	pSSC->SSC_PTCR = AT91C_PDC_TXTDIS;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_DisableRx
//* \brief Disable receive
//*----------------------------------------------------------------------------
 void AT91F_PDC_DisableRx (
	AT91PS_SSC pSSC )       // \arg pointer to a PDC controller
{
	pSSC->SSC_PTCR = AT91C_PDC_RXTDIS;
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_IsTxEmpty
//* \brief Test if the current transfer descriptor has been sent
//*----------------------------------------------------------------------------
 int AT91F_PDC_IsTxEmpty ( // \return return 1 if transfer is complete
	AT91PS_SSC pSSC )       // \arg pointer to a PDC controller
{
	return !(pSSC->SSC_TCR);
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_IsNextTxEmpty
//* \brief Test if the next transfer descriptor has been moved to the current td
//*----------------------------------------------------------------------------
 int AT91F_PDC_IsNextTxEmpty ( // \return return 1 if transfer is complete
	AT91PS_SSC pSSC )       // \arg pointer to a PDC controller
{
	return !(pSSC->SSC_TNCR);
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_IsRxEmpty
//* \brief Test if the current transfer descriptor has been filled
//*----------------------------------------------------------------------------
 int AT91F_PDC_IsRxEmpty ( // \return return 1 if transfer is complete
	AT91PS_SSC pSSC )       // \arg pointer to a PDC controller
{
	return !(pSSC->SSC_RCR);
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_IsNextRxEmpty
//* \brief Test if the next transfer descriptor has been moved to the current td
//*----------------------------------------------------------------------------
 int AT91F_PDC_IsNextRxEmpty ( // \return return 1 if transfer is complete
	AT91PS_SSC pSSC )       // \arg pointer to a PDC controller
{
	return !(pSSC->SSC_RNCR);
}
void AT91F_PDC_Rx_Close(AT91PS_SSC pSSC)       // \arg pointer to a PDC controller
{
	AT91F_PDC_DisableRx(pSSC);
	AT91F_PDC_SetNextRx(pSSC, (char *) 0, 0);
	AT91F_PDC_SetRx(pSSC, (char *) 0, 0);
}
void AT91F_PDC_Rx_Init(AT91PS_SSC pSSC)       // \arg pointer to a PDC controller
{
	AT91F_PDC_Rx_Close(pSSC);
	AT91F_PDC_EnableRx(pSSC);
}
void AT91F_PDC_Tx_Close(AT91PS_SSC pSSC){       // \arg pointer to a PDC controller
	AT91F_PDC_DisableTx(pSSC);
	AT91F_PDC_SetNextTx(pSSC, (char *) 0, 0);
	AT91F_PDC_SetTx(pSSC, (char *) 0, 0);
}
void AT91F_PDC_Tx_Init(AT91PS_SSC pSSC)       // \arg pointer to a PDC controller
{
	AT91F_PDC_Tx_Close(pSSC);
	AT91F_PDC_EnableTx(pSSC);
}
//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_Open
//* \brief Open PDC: disable TX and RX reset transfer descriptors, re-enable RX and TX
//*----------------------------------------------------------------------------
 void AT91F_PDC_Open (
	AT91PS_SSC pSSC)       // \arg pointer to a PDC controller
{
    //* Disable the RX and TX PDC transfer requests
	AT91F_PDC_DisableRx(pSSC);
	AT91F_PDC_DisableTx(pSSC);

	//* Reset all Counter register Next buffer first
	AT91F_PDC_SetNextTx(pSSC, (char *) 0, 0);
	AT91F_PDC_SetNextRx(pSSC, (char *) 0, 0);
	AT91F_PDC_SetTx(pSSC, (char *) 0, 0);
	AT91F_PDC_SetRx(pSSC, (char *) 0, 0);

    //* Enable the RX and TX SSC transfer requests
	AT91F_PDC_EnableRx(pSSC);
	AT91F_PDC_EnableTx(pSSC);
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_Close
//* \brief Close PDC: disable TX and RX reset transfer descriptors
//*---------------------------------------------------------------------------
 void AT91F_PDC_Close (
	AT91PS_SSC pSSC)       // \arg pointer to a PDC controller
{
    //* Disable the RX and TX PDC transfer requests
	AT91F_PDC_DisableRx(pSSC);
	AT91F_PDC_DisableTx(pSSC);

	//* Reset all Counter register Next buffer first
	AT91F_PDC_SetNextTx(pSSC, (char *) 0, 0);
	AT91F_PDC_SetNextRx(pSSC, (char *) 0, 0);
	AT91F_PDC_SetTx(pSSC, (char *) 0, 0);
	AT91F_PDC_SetRx(pSSC, (char *) 0, 0);

}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_SendFrame
//* \brief Close PDC: disable TX and RX reset transfer descriptors
//*----------------------------------------------------------------------------
 unsigned int AT91F_PDC_SendFrame(
	AT91PS_SSC pSSC,
	char *pBuffer,
	unsigned int szBuffer,
	char *pNextBuffer,
	unsigned int szNextBuffer )
{
	if (AT91F_PDC_IsTxEmpty(pSSC)) {
		//* Buffer and next buffer can be initialized
		AT91F_PDC_SetTx(pSSC, pBuffer, szBuffer);
		AT91F_PDC_SetNextTx(pSSC, pNextBuffer, szNextBuffer);
		return 2;
	}
	else if (AT91F_PDC_IsNextTxEmpty(pSSC)) {
		//* Only one buffer can be initialized
		AT91F_PDC_SetNextTx(pSSC, pBuffer, szBuffer);
		return 1;
	}
	else {
		//* All buffer are in use...
		return 0;
	}
}

//*----------------------------------------------------------------------------
//* \fn    AT91F_PDC_ReceiveFrame
//* \brief Close PDC: disable TX and RX reset transfer descriptors
//*----------------------------------------------------------------------------
 unsigned int AT91F_PDC_ReceiveFrame (
	AT91PS_SSC pSSC,
	char *pBuffer,
	unsigned int szBuffer,
	char *pNextBuffer,
	unsigned int szNextBuffer )
{
	if (AT91F_PDC_IsRxEmpty(pSSC)) {
		//* Buffer and next buffer can be initialized
		AT91F_PDC_SetRx(pSSC, pBuffer, szBuffer);
		AT91F_PDC_SetNextRx(pSSC, pNextBuffer, szNextBuffer);
		return 2;
	}
	else if (AT91F_PDC_IsNextRxEmpty(pSSC)) {
		//* Only one buffer can be initialized
		AT91F_PDC_SetNextRx(pSSC, pBuffer, szBuffer);
		return 1;
	}
	else {
		//* All buffer are in use...
		return 0;
	}
}

#define CHUNK_SIZE 160          // 20ms
#define SILENCE_CHUNK_SIZE 8    // 1ms 

static void AT91F_SSC_SetBaudrate (AT91PS_SSC pSSC, unsigned int mainClock,unsigned int speed){
        unsigned int baud_value;
        //* Define the baud rate divisor register
        if (speed == 0)
           baud_value = 0;
        else
        {
          unsigned int baud_value;
        //* Define the baud rate divisor register
        if (speed == 0)
           baud_value = 0;
        else
        {
           baud_value = (unsigned int) (mainClock * 10)/(2*speed);
           if ((baud_value % 10) >= 5)
                  baud_value = (baud_value / 10) + 1;
           else
                  baud_value /= 10;
        }
        pSSC->SSC_CMR = baud_value;
        }
        pSSC->SSC_CMR = baud_value;
}
static int AT91F_SSC_Configure (AT91PS_SSC pSSC, unsigned int syst_clock, unsigned int baud_rate, unsigned int clock_rx, unsigned int mode_rx, unsigned int clock_tx, unsigned int mode_tx, unsigned int* pFsos){
  //	DBG("SYS_CLOCK: %x BR: %x\n", syst_clock, baud_rate);
  	pSSC->SSC_IDR = (unsigned int) -1;
	pSSC->SSC_CR = (AT91C_SSC_SWRST | AT91C_SSC_RXDIS | AT91C_SSC_TXDIS);
	AT91F_SSC_SetBaudrate(pSSC, syst_clock, baud_rate);
	pSSC->SSC_RCMR =  clock_rx;
	pSSC->SSC_TCMR =  clock_tx;
	pSSC->SSC_RFMR =  mode_rx;
	pSSC->SSC_TFMR =  mode_tx;
	AT91F_PDC_Open(pSSC);
	*pFsos = mode_tx & AT91C_SSC_FSOS;
	//	DBG("\r\n\r\nCMR: %x RCMR: %x TCMR: %x RFMR: %x TFMR: %x fsos: %x\r\n", pSSC->SSC_CMR, clock_rx, mode_rx, clock_tx, mode_tx, *pFsos);
	return(0);
}
inline void AT91F_SSC_IF_Disable(AT91PS_SSC pSSC){
     pSSC->SSC_TFMR &= ~AT91C_SSC_FSOS;
      pSSC->SSC_TCMR &= ~AT91C_SSC_CKO;
      //DBG("\r\nIF Disable TCMR: %x  TFMR: %x\r\n",pSSC->SSC_TCMR, pSSC->SSC_TFMR);
}
inline void AT91F_SSC_IF_Enable(AT91PS_SSC pSSC, uint fsos){
      pSSC->SSC_TFMR |= fsos;
      pSSC->SSC_TCMR |= AT91C_SSC_CKO_CONTINOUS;
      //DBG("\r\nIF Enable TCMR: %x  TFMR: %x fsos: %x\r\n",pSSC->SSC_TCMR, pSSC->SSC_TFMR, fsos);
}
inline  void AT91F_SSC_EnableRx (AT91PS_SSC pSSC){
    pSSC->SSC_CR = AT91C_SSC_RXEN;
}
inline void AT91F_SSC_DisableRx (AT91PS_SSC pSSC){
    pSSC->SSC_CR = AT91C_SSC_RXDIS;
}
inline void AT91F_SSC_EnableTx (AT91PS_SSC pSSC){
    pSSC->SSC_CR = AT91C_SSC_TXEN;
}
inline void AT91F_SSC_DisableTx (AT91PS_SSC pSSC){
    pSSC->SSC_CR = AT91C_SSC_TXDIS;
}
inline void AT91F_SSC_EnableIt (AT91PS_SSC pSSC,unsigned int flag){
	pSSC->SSC_IER = flag;
}
inline void AT91F_SSC_DisableIt (AT91PS_SSC pSSC,unsigned int flag){
	pSSC->SSC_IDR = flag;
}
inline int AT91F_SSC_IsInterruptMasked(AT91PS_SSC pSSC, unsigned int flag){
        return (pSSC->SSC_IMR & flag);
}
inline int AT91F_AIC_IsInterruptMasked(unsigned int flag){
        return (AT91_SYS->AIC_IMR & flag);
}
inline int AT91F_AIC_IsInterruptPending(unsigned int flag){
        return (AT91_SYS->AIC_IPR & flag);
}
inline int AT91F_SSC_GetStatus(AT91PS_SSC pSSC){
        return (pSSC->SSC_SR);
}


#define MAX_SSC_NB 1

static struct hdw_info{
  int dev;
  int used;
  int lock;
  int irq;
  AT91PS_SSC ssc;
  unsigned int fsos;   // last frame sync settings 
  int sclk;
  int bits;
}hinfos[MAX_SSC_NB] =  {{0}};

static int configure(struct hdw_info *hi, int speed, int bits)
{
  if(speed < 0 || bits<0) 
    return -1;

  /* Configure codec interface as Master 
    Receive clock is Master Clock divided and is output,
    Transmit clock is taken from Receive Clock,
     */
  if(speed)
      hi->sclk = speed;
  if(bits)
      hi->bits = bits-1;
  if(hi->bits > SSC_BITS-1)
    hi->bits = SSC_BITS-1;
  return( AT91F_SSC_Configure(hi->ssc, AT91C_MASTER_CLOCK, hi->sclk, RCMR_VAL, RFMR_VAL|hi->bits, TCMR_VAL, TFMR_VAL|hi->bits, &(hi->fsos)));
}
#define AT91C_VA_BASE_PDC_SSC1	AT91_IO_P2V(AT91C_BASE_PDC_SSC1)

static void hdw_init(int hidx)
{
  struct hdw_info* hi = &hinfos[hidx];
  if(hidx == 0){
    hi->ssc = (AT91PS_SSC) AT91C_VA_BASE_SSC1;
    hi->used = hi->lock = 0;
    hi->irq = AT91C_ID_SSC1;
    AT91_SYS->PIOB_ASR = 0xFC0;
    AT91_SYS->PIOB_BSR = 0;
    AT91_SYS->PIOB_PDR = 0xFC0; // Set in Periph mode
    AT91_SYS->PMC_PCER = 1<<15;
    configure(hi, SSC_SCLK, SSC_BITS);     /* configure with default values */
    AT91F_SSC_IF_Disable(hi->ssc);         // stop output SCLK and FSYNC
  }
}
static void hdw_close(struct hdw_info *hi){
  AT91F_PDC_Close (hi->ssc);
  AT91F_SSC_DisableIt (hi->ssc,0xffff);
  AT91F_SSC_DisableRx (hi->ssc);
  AT91F_SSC_DisableTx (hi->ssc);
  AT91F_SSC_IF_Disable(hi->ssc);
  // ports stay configured for codec
}
struct timeval inttime;

void at91ssc_interrupt(int irq, void *dev_id, struct pt_regs *regs){
  struct hdw_info* hi = dev_id;
  int status =  AT91F_SSC_GetStatus(hi->ssc); 
  //  do_gettimeofday(&inttime);
  if((status & AT91C_SSC_ENDTX) && (status & AT91C_SSC_TXENA) && AT91F_SSC_IsInterruptMasked(hi->ssc,AT91C_SSC_ENDTX)){
     AT91F_SSC_DisableIt(hi->ssc, AT91C_SSC_ENDTX);
     DMAbuf_outputintr(hi->dev, 0);
  }
  if((status & AT91C_SSC_ENDRX) && (status & AT91C_SSC_RXENA) && AT91F_SSC_IsInterruptMasked(hi->ssc,AT91C_SSC_ENDRX)){
    AT91F_SSC_DisableIt(hi->ssc, AT91C_SSC_ENDRX);
    DMAbuf_inputintr(hi->dev);
  }
}

int at91ssc_ioctl(int dev, unsigned int cmd, caddr_t arg){
  DBG("Ioctl %x, %x\n",cmd, *(ulong*)arg);
  return 0;
}
void at91ssc_output (int dev, unsigned long buf, int count, int intrflag){
  struct hdw_info* hi = audio_devs[dev]->devc; 
  int status =  AT91F_SSC_GetStatus(hi->ssc); 
  spin_lock_irq(&hi->lock);
  if(AT91F_PDC_SendFrame (hi->ssc, (char *)buf,count/2,0,0)){
    if(!(status & AT91C_SSC_TXENA)){
	 AT91F_PDC_EnableTx ( hi->ssc);
	 AT91F_SSC_EnableTx ( hi->ssc);
    }
    AT91F_SSC_EnableIt(hi->ssc, AT91C_SSC_ENDTX);
    spin_unlock_irq(&hi->lock);
  }else{
    spin_unlock_irq(&hi->lock);
    DBG("Transmitted Data Lost !!!\n");
  }
}
void at91ssc_start_input(int dev, unsigned long buf, int count, int intrflag){
  struct hdw_info* hi = audio_devs[dev]->devc;
  int status =  AT91F_SSC_GetStatus(hi->ssc);
  struct timeval mytime;
  //    DBG("Read %x %d\n",buf, count);
#if 0
    do_gettimeofday(&mytime);
    long ds,dus;

    ds = mytime.tv_sec - inttime.tv_sec;
    dus = mytime.tv_usec - inttime.tv_usec;
    if (dus < 0)
      {
	ds -= 1;
	dus += 1000000;
      }


    DBG("Delay: %d.%d\n",ds,dus);
#endif
    spin_lock_irq(&hi->lock);
   if(AT91F_PDC_ReceiveFrame (hi->ssc, (char *)buf,count/2,0,0)){
     if(!(status & AT91C_SSC_RXENA)){
       //         DBG("Enable ENDRX Ints %x %d\n",buf, count);
	 AT91F_PDC_EnableRx ( hi->ssc);
	 AT91F_SSC_EnableRx ( hi->ssc);
     }
     AT91F_SSC_EnableIt(hi->ssc, AT91C_SSC_ENDRX);
     spin_unlock_irq(&hi->lock);
   }else{
     spin_unlock_irq(&hi->lock);
     DBG("Received Data Lost !!!\n");
   }
}
int at91ssc_prepare_for_input(int dev, int bufsize, int nbufs){
  struct hdw_info* hi = audio_devs[dev]->devc;
  DBG("Prepare read %d, %d\n",bufsize, nbufs);
  audio_devs[dev]->dmap_in->flags |= DMA_NODMA;
  AT91F_PDC_Rx_Init(hi->ssc);
  return 0;
}
int at91ssc_prepare_for_output(int dev, int bufsize, int nbufs){
  struct hdw_info* hi = audio_devs[dev]->devc;
  DBG("Prepare write %d, %d\n",bufsize, nbufs);
  audio_devs[dev]->dmap_out->flags |= DMA_NODMA;
  AT91F_PDC_Tx_Init(hi->ssc);
  return 0;
}
void at91ssc_halt_io(int dev){
  struct hdw_info* hi = audio_devs[dev]->devc;
  int status =  AT91F_SSC_GetStatus(hi->ssc); 
  DBG("HALT when SSC status %x\n",status);
  AT91F_SSC_DisableIt(hi->ssc, AT91C_SSC_ENDRX | AT91C_SSC_ENDTX);
  AT91F_PDC_Rx_Close(hi->ssc);
  AT91F_PDC_Tx_Close(hi->ssc);
  AT91F_PDC_DisableTx(hi->ssc);
  AT91F_PDC_DisableRx(hi->ssc);
}
void at91ssc_copy_from_user(int dev, char *localbuf, int localoffs,
                        const char *userbuf, int useroffs,
                        int max_in, int max_out, int *used, int *returned, int len){
  DBG("Copy lb %x loff %x ub %x uoff %x \n  maxin %d maxout %d len %d\n", localbuf, localoffs, userbuf, useroffs, max_in, max_out, len);
}
void at91ssc_halt_input(int dev){
  struct hdw_info* hi = audio_devs[dev]->devc; 
  int status =  AT91F_SSC_GetStatus(hi->ssc); 
  DBG("HALT IN when SSC status %x\n",status);
  AT91F_SSC_DisableIt(hi->ssc, AT91C_SSC_ENDRX);
  AT91F_PDC_Rx_Close(hi->ssc);
}
void at91ssc_halt_output(int dev){
  struct hdw_info* hi = audio_devs[dev]->devc;
  int status =  AT91F_SSC_GetStatus(hi->ssc); 
  DBG("HALT OUT when SSC status %x\n",status);
  AT91F_SSC_DisableIt(hi->ssc, AT91C_SSC_ENDTX);
  AT91F_PDC_Tx_Close(hi->ssc);
}
void at91ssc_trigger(int dev, int bits){
}
int at91ssc_set_speed(int dev, int speed){
  return FSYNC;
}
unsigned int at91ssc_set_bits(int dev, unsigned int bits){
  return SSC_BITS;
}
ushort at91ssc_set_channels(int dev, short channels){
  return 1;
}
void postprocess_write(int dev){
  DBG("PP write \n");
}
void preprocess_read(int dev){
  DBG("PP read \n");
}
void at91ssc_mmap(int dev){
  DBG("MMAP \n");
}
int at91ssc_open (int dev, int mode){
  struct hdw_info* hi = audio_devs[dev]->devc;
  DBG("Open mode %d \n",mode);
  audio_devs[dev]->dmap_out->flags |= DMA_NODMA;
  audio_devs[dev]->dmap_in->flags |= DMA_NODMA;

  if(hi->used)
    return 0;  // allow multiple open for test,(instead of -EBUSY)
  hi->used++;
  AT91F_SSC_IF_Enable(hi->ssc, hi->fsos);
  AT91F_PDC_Open(hi->ssc);
  DBG("Ready %d \n",mode);
  return 0;
}
void at91ssc_close (int dev){
  struct hdw_info* hi = audio_devs[dev]->devc;
  DBG("Close owners:%d\n",hi->used);
  if(hi->used-- > 1)
    return;
  // all apps have gone, close codec interface
  at91ssc_halt_io(dev);
  hdw_close(hi);  
}

char *at91ssc_name = "AT91RM9200 SSC audio driver";

static struct audio_driver at91ssc_audio_driver =	/* AT91SSC */
{
	owner:		THIS_MODULE,
	open:		at91ssc_open,
	close:		at91ssc_close,
	output_block:	at91ssc_output,
	start_input:	at91ssc_start_input,
	ioctl:          at91ssc_ioctl,
	prepare_for_input:	at91ssc_prepare_for_input,
	prepare_for_output:	at91ssc_prepare_for_output,
	halt_io:	at91ssc_halt_io,
	halt_input:     at91ssc_halt_input,
	halt_output:    at91ssc_halt_output,
	trigger:	at91ssc_trigger,
	set_speed:	at91ssc_set_speed,
	set_bits:	at91ssc_set_bits,
	mmap:		at91ssc_mmap
};
//	copy_user:	at91ssc_copy_from_user,

static void init(void){
  struct hdw_info* hi = &hinfos[0];
    hdw_init(0);
    printk("AT91SSC: SSC1 initializing\n");
    /* register as audio device */
   hi->dev = sound_install_audiodrv(AUDIO_DRIVER_VERSION, at91ssc_name, &at91ssc_audio_driver,sizeof(struct audio_driver),
				    DMA_DUPLEX | DMA_AUTOMODE , AFMT_S16_LE, hi, -1, -1);

  if(request_irq(AT91C_ID_SSC1, at91ssc_interrupt, 0, audio_devs[hi->dev]->name, hi)<0){
	printk(KERN_ERR "SB: Can't allocate IRQ%d\n", hi->irq);
  }else{
    printk("AT91SSC: Driver installed as audio device %d \n", hi->dev);
  }
  audio_devs[hi->dev]->max_fragment = 9; 

} 

static void remove(void){
  struct hdw_info* hi = &hinfos[0];
  hdw_close(hi);
  hi->used = hi->lock = 0;
  free_irq(AT91C_ID_SSC1, hi);
}
static int __init init_at91ssc(void){
  init();
  return 0;
}
static void __exit cleanup_at91ssc(void){
  remove();
}

module_init(init_at91ssc);
module_exit(cleanup_at91ssc);


#ifndef MODULE

static int __init at91ssc_setup(char *options)
{
  init();
  return 0;
}
__setup("at91rm9200ssc=", at91ssc_setup);
#endif

